# pylint: skip-file
import os
import glob
import argparse
import pandas as pd
from petpal.kinetic_modeling import graphical_analysis,rtm_analysis
from petpal.preproc import image_operations_4d, motion_corr, register, segmentation_tools
from petpal.preproc import symmetric_geometric_transfer_matrix as sgtm
import petpal.preproc.regional_tac_extraction
import petpal.preproc.standard_uptake_value
from petpal.utils.bids_utils import gen_bids_like_dir_path, gen_bids_like_filename, gen_bids_like_filepath
from petpal.utils.image_io import km_regional_fits_to_tsv
from petpal.utils import useful_functions
from petpal.utils.constants import HALF_LIVES


_VAT_EXAMPLE_ = (r"""
Example:
  - Running many subjects:
    petpal-vat-proc --subjects participants.tsv --out-dir /path/to/output --pet-dir /path/to/pet/folder/ --reg-dir /path/to/subject/Registrations/
""")
def infer_group(sub_id: str):
    """
    Infer the group a subject belongs to based on the subject ID.

    PIB and VATNL belong to 'HC' group while VATDYS belongs to CD group.
    """
    group = 'UNK'
    if 'PIB' in sub_id:
        group = 'HC'
    if 'VATNL' in sub_id:
        group = 'HC'
    if 'VATDYS' in sub_id:
        group = 'CD'
    return group


def rename_subs(sub: str):
    """
    Handle converting subject ID to BIDS structure.

    VATDYS0XX -> sub-VATDYS0XX
    PIBXX-YYY_VYrZ -> sub-PIBXXYYY_ses-VYrZ

    returns:
        - subject part string
        - session part string
    """
    if 'VAT' in sub:
        return [f'sub-{sub}', 'ses-VYr0']
    elif 'PIB' in sub:
        subname, sesname = sub.split('_')
        subname = subname.replace('-','')
        subname = f'sub-{subname}'
        sesname = f'ses-{sesname}'
        return [subname, sesname]


def vat_protocol(subjstring: str,
                 out_dir: str,
                 pet_dir: str,
                 fs_dir: str,
                 skip: list):
    sub, ses = rename_subs(subjstring)
    sub_id = sub.replace('sub-','')
    ses_id = ses.replace('ses-','')
    segmentation_label_file = '/home/usr/goldmann/dseg.tsv'
    motion_target = (0,600)
    reg_pars = {'aff_metric': 'mattes','type_of_transform': 'DenseRigid'}
    half_life = HALF_LIVES['f18']
    suvr_start = 1800
    suvr_end = 7200
    pvc_fwhm_mm = 4.2
    if 'VAT' in sub:
        pet_file = f'{pet_dir}/{sub}/pet/{sub}_pet.nii.gz'
        freesurfer_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/aparc+aseg.mgz')[0]
        brainstem_segmentation_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/brainstemSsLabels.v13.FSvoxelSpace.mgz')[0]
        mprage_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/nu.mgz')[0]
        atlas_warp_file = f'PLACEHOLDER'
        mpr_brain_mask_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/brainmask.mgz')[0]
    else:
        pet_file = f'{pet_dir}/{sub}/{ses}/pet/{sub}_{ses}_trc-18FVAT_pet.nii.gz'
        freesurfer_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/aparc+aseg.mgz')[0]
        brainstem_segmentation_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/brainstemSsLabels.v13.FSvoxelSpace.mgz')[0]
        mprage_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/nu.mgz')[0]
        atlas_warp_file = f'PLACEHOLDER'
        mpr_brain_mask_file = glob.glob(f'{fs_dir}/{subjstring}*/mri/brainmask.mgz')[0]
    real_files = [
        pet_file,
        freesurfer_file,
        mprage_file,
        brainstem_segmentation_file,
        atlas_warp_file,
        mpr_brain_mask_file
    ]
    for check in real_files:
        if not os.path.exists(check):
            print(f'{check} not found')
    print(real_files)


    def vat_bids_filepath(suffix,folder,**extra_desc):

        new_dir = gen_bids_like_dir_path(sub_id=sub_id,
                                         ses_id=ses_id,
                                         sup_dir=out_dir,
                                         modality=folder)
        os.makedirs(new_dir,exist_ok=True)
        new_file = gen_bids_like_filepath(sub_id=sub_id,
                                          ses_id=ses_id,
                                          bids_dir=out_dir,
                                          modality=folder,
                                          suffix=suffix,
                                          **extra_desc)

        return new_file


    # preprocessing
    pet_cropped_file = vat_bids_filepath(suffix='pet',folder='pet',crop='003')
    if 'crop' not in skip:
        image_operations_4d.SimpleAutoImageCropper(input_image_path=pet_file,
                                                   out_image_path=pet_cropped_file,
                                                   thresh_val=0.03)

    pet_moco_file = vat_bids_filepath(suffix='pet',folder='pet',moco='windowed')
    if 'moco' not in skip:
        motion_corr.windowed_motion_corr_to_target(input_image_path=pet_cropped_file,
                                                out_image_path=pet_moco_file,
                                                motion_target_option=motion_target,
                                                w_size=300)

    pet_reg_anat_file = vat_bids_filepath(suffix='pet',folder='pet',moco='windowed',space='mpr')
    if 'register' not in skip:
        register.register_pet(input_reg_image_path=pet_moco_file,
                            out_image_path=pet_reg_anat_file,
                            reference_image_path=mprage_file,
                            motion_target_option=motion_target,
                            half_life=half_life,
                            verbose=True,
                            **reg_pars)
    
    vat_wm_ref_region_roi_file = vat_bids_filepath(suffix='seg',folder='pet',desc='RefRegionROI')
    vat_wm_ref_segmentation_file = vat_bids_filepath(suffix='seg',folder='pet',desc='RefRegionSegmentation')
    if 'refregion' not in skip:
        segmentation_tools.vat_wm_ref_region(input_segmentation_path=freesurfer_file,
                                             out_segmentation_path=vat_wm_ref_region_roi_file)
        segmentation_tools.vat_wm_region_merge(wmparc_segmentation_path=freesurfer_file,
                                               bs_segmentation_path=brainstem_segmentation_file,
                                               wm_ref_segmentation_path=vat_wm_ref_region_roi_file,
                                               out_image_path=vat_wm_ref_segmentation_file)

    tac_save_dir = gen_bids_like_dir_path(sub_id=sub_id,ses_id=ses_id,sup_dir=out_dir,modality='tacs')
    os.makedirs(tac_save_dir,exist_ok=True)
    tac_prefix = f'{sub}_{ses}'
    if 'tacs' not in skip:
        petpal.preproc.regional_tac_extraction.write_tacs(input_image_path=pet_reg_anat_file,
                                                          label_map_path=segmentation_label_file,
                                                          segmentation_image_path=vat_wm_ref_segmentation_file,
                                                          out_tac_dir=tac_save_dir,
                                                          verbose=True,
                                                          out_tac_prefix=tac_prefix)

    # kinetic modeling
    wmref_tac_path = vat_bids_filepath(suffix='tac',folder='tacs',seg='WMRef',ext='.tsv')
    mrtm_save_dir = gen_bids_like_dir_path(sub_id=sub_id,ses_id=ses_id,sup_dir=out_dir,modality='mrtm_fits')
    km_save_dir = gen_bids_like_dir_path(sub_id=sub_id,ses_id=ses_id,sup_dir=out_dir,modality='km')
    os.makedirs(km_save_dir,exist_ok=True)
    mrtm_save_path = vat_bids_filepath(suffix='fits',model='mrtm1',folder='km',ext='.tsv')
    os.makedirs(mrtm_save_dir,exist_ok=True)
    mrtm1_path = gen_bids_like_filename(sub_id=sub_id,ses_id=ses_id,model='mrtm1',suffix='fits',ext='')
    if 'mrtm1' not in skip:
        mrtm1_analysis = rtm_analysis.MultiTACRTMAnalysis(ref_tac_path=wmref_tac_path,
                                                          roi_tacs_dir=tac_save_dir,
                                                          output_directory=mrtm_save_dir,
                                                          output_filename_prefix=mrtm1_path,
                                                          method='mrtm')
        mrtm1_analysis.run_analysis(t_thresh_in_mins=10)
        mrtm1_analysis.save_analysis()
        km_regional_fits_to_tsv(fit_results_dir=mrtm_save_dir,out_tsv_dir=mrtm_save_path)

    logan_save_dir = gen_bids_like_dir_path(sub_id=sub_id,ses_id=ses_id,sup_dir=out_dir,modality='logan_fits')
    logan_save_path = vat_bids_filepath(suffix='fits',model='altlogan',folder='km',ext='.tsv')
    os.makedirs(logan_save_dir,exist_ok=True)
    logan_path = gen_bids_like_filename(sub_id=sub_id,ses_id=ses_id,model='altlogan',suffix='fits',ext='')
    if 'logan' not in skip:
        graphical_model = graphical_analysis.MultiTACGraphicalAnalysis(
            input_tac_path=wmref_tac_path,
            roi_tacs_dir=tac_save_dir,
            output_directory=logan_save_dir,
            output_filename_prefix=logan_path,
            method='alt_logan',
            fit_thresh_in_mins=10
        )
        graphical_model.run_analysis()
        graphical_model.save_analysis()
        km_regional_fits_to_tsv(fit_results_dir=logan_save_dir,out_tsv_dir=logan_save_path)

    # suvr
    wss_file_path = vat_bids_filepath(suffix='pet',folder='pet',space='mpr',desc='WSS')
    suvr_file_path = vat_bids_filepath(suffix='pet',folder='pet',space='mpr',desc='SUVR')
    if 'suvr' not in skip:
        useful_functions.weighted_series_sum(input_image_path=pet_reg_anat_file,
                                             half_life=half_life,
                                             verbose=True,
                                             start_time=suvr_start,
                                             end_time=suvr_end,
                                             out_image_path=wss_file_path)
        petpal.preproc.standard_uptake_value.suvr(input_image_path=wss_file_path,
                                 out_image_path=suvr_file_path,
                                 segmentation_image_path=vat_wm_ref_segmentation_file,
                                 ref_region=1,
                                 verbose=True)

    if 'pvc' not in skip:
        suvr_pvc_path = vat_bids_filepath(suffix='pet',folder='pet',space='mpr',pvc='SGTM',desc='SUVR',ext='.tsv')
        sgtm.Sgtm(input_image_path=suvr_file_path,
                  segmentation_image_path=vat_wm_ref_segmentation_file,
                  fwhm=pvc_fwhm_mm,
                  out_tsv_path=suvr_pvc_path)


def main():
    """
    VAT command line interface
    """
    parser = argparse.ArgumentParser(prog='petpal-vat-proc',
                                     description='Command line interface for running VAT processing.',
                                     epilog=_VAT_EXAMPLE_, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-s','--subjects',required=True,help='Path to participants.tsv')
    parser.add_argument('-o','--out-dir',required=True,help='Output directory analyses are saved to.')
    parser.add_argument('--vatdys-pet',required=True,help='Path to BIDS directory storing VATDYS PET')
    parser.add_argument('--vatdys-fs',required=True,help='Path to directory storing VATDYS FreeSurfer')
    parser.add_argument('--vatnl-pet',required=True,help='Path to BIDS directory storing VATNL PET')
    parser.add_argument('--vatnl-fs',required=True,help='Path to directory storing VATNL FreeSurfer')
    parser.add_argument('--pib-pet',required=True,help='Path to BIDS directory storing PIB PET')
    parser.add_argument('--pib-fs',required=True,help='Path to directory storing PIB FreeSurfer')

    parser.add_argument('--skip',required=False,help='List of steps to skip',nargs='+',default=[])
    args = parser.parse_args()


    subs_sheet = pd.read_csv(args.subjects,sep='\t')
    subs = subs_sheet['participant_id']

    for sub in subs:
        if sub[:6]=='VATDYS':
            pet_dir = args.vatdys_pet
            fs_dir = args.vatdys_fs
        elif sub[:5]=='VATNL':
            pet_dir = args.vatnl_pet
            fs_dir = args.vatnl_fs
        elif sub[:3]=='PIB':
            pet_dir = args.pib_pet
            fs_dir = args.pib_fs
        try:
            vat_protocol(sub,args.out_dir,pet_dir,fs_dir,skip=args.skip)
        except:
            print(f'Running subject {sub} failed; trying next one.')
main()



r"""
This module provides classes and utilities for computing Image-Derived Input Functions
(IDIF) using a PCA-guided approach. It is designed to integrate into preprocessing pipelines
and offers functionality for optimizing voxel selection, applying fitting techniques,
and selecting top-ranked voxels from PCA components.

Classes:
--------

1. :class:`PCAGuidedIDIFMixin`:
    - Acts as a mixin to provide shared functionality across PCA-guided steps.
    - Manages input/output paths, initializes configurations, and helps connect pipeline steps.
    - Provides methods to infer output paths and set input paths dynamically.

2. :class:`PCAGuidedFitIDIFStep`:
    - Represents a pipeline step for optimizing voxel selection to fit the IDIF using PCA-guided methods.
    - Extends both :class:`~.ObjectBasedStep` and :class:`~.PCAGuidedIDIFMixin` to integrate optimization fitting into a pipeline.
    - Includes parameters to control optimization behavior, such as:
        - **alpha**: Weight for the peak term to encourage a high TAC peak.
        - **beta**: Weight for the smoothness term to ensure low total variation.
        - **method**: The optimization method for fitting.

3. :class:`PCAGuidedTopVoxelsIDIFStep`:
    - Represents a pipeline step to calculate the IDIF by selecting the top `N` voxels from a specific PCA component.
    - Extends both :class:`~.ObjectBasedStep` and :class:`~.PCAGuidedIDIFMixin` for pipeline functionality.
    - Allows customizing the selected PCA component and the number of voxels to include.

Key Features:
-------------
- **Pipeline Integration**:
    All classes are designed to be modular and compatible with preprocessing pipelines.
    They inherit functionalities from :class:`~.ObjectBasedStep` and :class:`~.StepsAPI`.

- **BIDS-Compatible Output**:
    Output paths for TAC files are generated in a BIDS-like format using subject and session information
    extracted from the input image paths.

- **Optimization and Voxel Selection**:
    Classes include functionality to:
        - Select voxels based on PCA component ranking.
        - Optimize voxel selection iteratively for better IDIF calculation.
        - Use terms like smoothness, noise reduction, and peak emphasis for guiding optimization.

"""

import warnings

from ..input_function import pca_guided_idif
from ..pipelines.preproc_steps import ImageToImageStep
from ..pipelines.steps_base import StepsAPI, ObjectBasedStep
from ..utils.bids_utils import gen_bids_like_filepath, parse_path_to_get_subject_and_session_id, snake_to_camel_case


class PCAGuidedIDIFMixin(StepsAPI):
    r"""Mixin class to provide shared functionality for PCA-guided IDIF pipeline steps.

    This class is designed to standardize and extend functionality across various PCA-guided
    IDIF calculation steps. It handles input and output path management, configuration initialization,
    and helps define the relationships between preprocessing pipeline steps.

    Inheritance:
        Inherits from the :class:`~.StepsAPI` base class.

    Attributes:
        init_kwargs (dict): A dictionary containing initial configuration parameters.
        call_kwargs (dict): A dictionary containing runtime call-specific parameters.
        input_image_path (str): Path to the input 4D-PET image.
        mask_image_path (str): Path to the mask image in the same space as 4D-PET image.
        output_tac_path (str): Path where output TAC data will be saved.
        num_pca_components (int): Number of PCA components to compute.
        verbose (bool): Whether to show detailed diagnostic output or not.
        name (str): Name of the pipeline step (default is "pca_guided_idif_base").

    """
    def __init__(self,
                 input_image_path: str,
                 mask_image_path: str,
                 output_tac_path: str,
                 num_pca_components: int,
                 verbose: bool
                 ):
        StepsAPI.__init__(self)
        self.init_kwargs = {'input_image_path': input_image_path,
                            'mask_image_path': mask_image_path,
                            'output_tac_path': output_tac_path,
                            'num_pca_components': num_pca_components,
                            'verbose': verbose
                            }
        self.call_kwargs = {}
        self.input_image_path = input_image_path
        self.mask_image_path = mask_image_path
        self.output_tac_path = output_tac_path
        self.num_pca_components = num_pca_components
        self.verbose = verbose
        self.name = 'pca_guided_idif_base'


    @property
    def input_image_path(self):
        return self.init_kwargs['input_image_path']

    @input_image_path.setter
    def input_image_path(self, value):
        self.init_kwargs['input_image_path'] = value

    @property
    def mask_image_path(self):
        return self.init_kwargs['mask_image_path']

    @mask_image_path.setter
    def mask_image_path(self, value):
        self.init_kwargs['mask_image_path'] = value

    @property
    def output_tac_path(self) -> str:
        return self.init_kwargs['output_tac_path']

    @output_tac_path.setter
    def output_tac_path(self, value):
        self.init_kwargs['output_tac_path'] = value

    @property
    def num_pca_components(self):
        return self.init_kwargs['num_pca_components']

    @num_pca_components.setter
    def num_pca_components(self, value):
        self.init_kwargs['num_pca_components'] = value

    @property
    def verbose(self):
        return self.init_kwargs['verbose']

    @verbose.setter
    def verbose(self, value):
        self.init_kwargs['verbose'] = value

    def set_input_as_output_from(self, *sending_steps: ImageToImageStep) -> None:
        """Sets the input image paths based on the output paths from other steps in the pipeline.

        The first sending step will set the input image path, and the second sending step will
        set the second image path.

        Args:
            sending_steps (tuple[ImageToImageStep]): Two pipeline steps whose outputs will be used
                as the input image path and second image input path.

        Raises:
            AssertionError: If the number of provided sending steps is not exactly two.
        """
        assert len(sending_steps) == 2, "ImagePairToArrayStep must have 2 sending ImageToImageStep steps."
        if isinstance(sending_steps[0], ImageToImageStep):
            self.input_image_path = sending_steps[0].output_image_path
        else:
            super().set_input_as_output_from(sending_steps[0])
        if isinstance(sending_steps[1], ImageToImageStep):
            self.mask_image_path = sending_steps[1].output_image_path
        else:
            super().set_input_as_output_from(sending_steps[1])

    def infer_outputs_from_inputs(self,
                                  out_dir: str,
                                  der_type: str = 'tacs',
                                  suffix: str = 'tac',
                                  ext: str = '.tsv',
                                  **extra_desc):
        r"""Infers the output array path based on the inputs and specified parameters.

        This method generates a BIDS-like derivatives filepath for the output based on the subject and
        session IDs extracted from the input image path.

        Args:
            out_dir (str): Directory where the output array will be saved.
            der_type (str, optional): Type of derivative. Will set the sub-directory in `out_dir`. Defaults to 'tacs'.
            suffix (str, optional): Suffix for the output filename. Defaults to 'tac'.
            ext (str, optional): File extension for the output file. Defaults to '.tsv'.
            **extra_desc: Additional descriptive parameters for the output filename.

        """
        sub_id, ses_id = parse_path_to_get_subject_and_session_id(self.input_image_path)
        step_name_in_camel_case = snake_to_camel_case(self.name)
        filepath = gen_bids_like_filepath(sub_id=sub_id, ses_id=ses_id, suffix=suffix, bids_dir=out_dir,
                                          modality=der_type, ext=ext, desc=step_name_in_camel_case, **extra_desc)
        self.output_tac_path = filepath


class PCAGuidedFitIDIFStep(ObjectBasedStep, PCAGuidedIDIFMixin):
    r"""Pipeline step for fitting the PCA-guided IDIF using an optimization-based approach.

    This step extends the :class:`~.PCAGuidedIDIFMixin` for shared functionality and the :class:`~.ObjectBasedStep`
    for pipeline execution. It uses the :class:`~.PCAGuidedIdifFitter` class to iteratively optimize voxel
    selection for IDIF generation based on specified parameters.

    Inheritance:
        Inherits from :class:`~.ObjectBasedStep` and :class:`~.PCAGuidedIDIFMixin`.

    Attributes:
        alpha (float): Parameter weight for the peak term (emphasizes high TAC peaks).
        beta (float): Parameter weight for the smoothness term (encourages smoother signals).
        method (str): Optimization method to be used during fitting.

    Example:

        .. code-block:: python

            from petpal.pipelines.pca_guided_idif_steps import PCAGuidedFitIDIFStep

            step = PCAGuidedFitIDIFStep(input_image_path='/path/to/image.nii',
                                        mask_image_path='/path/to/mask.nii',
                                        output_array_path='/path/to/output.tsv',
                                        num_pca_components=5,
                                        verbose=True,
                                        alpha=0.5,
                                        beta=0.3,
                                        method='ampgo')
            ## Running and saving the results
            step()  # Executes the step

    """
    def __init__(self,
                 input_image_path: str,
                 mask_image_path: str,
                 output_array_path: str,
                 num_pca_components: int,
                 verbose: bool,
                 alpha: float,
                 beta: float,
                 method: str,
                 **meth_kwargs):
        r"""Initializes the PCA-guided fitting IDIF step.

        Args:
            input_image_path (str): Path to the input 4D-PET image.
            mask_image_path (str): Path to the mask image in the same space as 4D-PET image.
            output_array_path (str): Path where the resulting TAC data will be saved.
            num_pca_components (int): Number of PCA components to consider.
            verbose (bool): If `True`, enables detailed diagnostic output.
            alpha (float): Parameter weight for the peak term.
            beta (float): Parameter weight for the smoothness term.
            method (str): Optimization method to use for fitting.
            **meth_kwargs: Additional keyword arguments for the optimization method.

        Side Effects:
            - Initializes `python-guided-idif` fitter with the provided parameters.
            - Populates `call_kwargs` with runtime-specific options for the fitting process.

        """
        PCAGuidedIDIFMixin.__init__(self,
                                    input_image_path=input_image_path,
                                    mask_image_path=mask_image_path,
                                    output_tac_path=output_array_path,
                                    num_pca_components=num_pca_components,
                                    verbose=verbose)
        ObjectBasedStep.__init__(self,
                                 name='pca_guided_fit_idif',
                                 class_type=pca_guided_idif.PCAGuidedIdifFitter,
                                 init_kwargs={**self.init_kwargs},
                                 call_kwargs=(self.call_kwargs | {'alpha' : alpha,
                                                                  'beta'  : beta,
                                                                  'method': method,
                                                                  **meth_kwargs}), )

        self.alpha = alpha
        self.beta = beta
        self.method = method

    @property
    def alpha(self):
        return self.call_kwargs['alpha']

    @alpha.setter
    def alpha(self, value):
        self.call_kwargs['alpha'] = value

    @property
    def beta(self):
        return self.call_kwargs['beta']

    @beta.setter
    def beta(self, value):
        self.call_kwargs['beta'] = value

    @property
    def method(self):
        return self.call_kwargs['method']

    @method.setter
    def method(self, value):
        self.call_kwargs['method'] = value

    @classmethod
    def default_pca_guided_idif_fit(cls, name: str = 'pca_guided_fit_idif', verbose=False, **overrides):
        r"""Creates a default instance for PCA-guided IDIF calculation using an optimization-based fitting approach.


        Uses :class:`~.pca_guided_idif.PCAGuidedIdifFitter`

        All file paths are set to empty strings by default.

        Args:
            name (str): Name of the step. Defaults to 'pca_guided_fit_idif'.
            verbose (bool): Whether to enable verbose mode (default: False).
            **overrides: Parameter values to override the default settings.

        Returns:
            PCAGuidedFitIDIFStep: A new instance of the PCA-guided IDIF fitting step.

        Default Parameters:
            - input_image_path: '' (empty string)
            - mask_image_path: '' (empty string)
            - output_array_path: '' (empty string)
            - num_pca_components: 3
            - verbose: False
            - alpha: 2.5
            - beta: 0.0
            - method: 'dual_annealing'

        Side Effects:
            - If an invalid override is provided, a warning is issued, and the default instance is returned instead.

        """
        defaults = dict(input_image_path='',
                        mask_image_path='',
                        output_array_path='',
                        num_pca_components=3,
                        verbose=verbose,
                        alpha=2.5,
                        beta=0.0,
                        method='dual_annealing')

        override_dict = defaults | overrides

        try:
            out_class = cls(**override_dict)
        except RuntimeError as err:
            warnings.warn(f"Invalid override: {err}. Using default instance instead.", stacklevel=2)
            out_class = cls(**defaults)
        out_class.name = name
        return out_class

class PCAGuidedTopVoxelsIDIFStep(ObjectBasedStep, PCAGuidedIDIFMixin):
    r"""Pipeline step for PCA-guided IDIF calculation using top-ranked voxels.

    This step extends the :class:`~.PCAGuidedIDIFMixin` and provides functionality to select the top `N`
    voxels from a specified PCA component for IDIF generation. It uses the
    :class:`~.PCAGuidedTopVoxelsIDIF` class for its calculations.

    Inheritance:
        Inherits from :class:`~.ObjectBasedStep` and :class:`~.PCAGuidedIDIFMixin`.

    Attributes:
        selected_component (int): PCA component selected for voxel ranking.
        num_of_voxels (int): Number of top-ranking voxels to select.

    Example:

        .. code-block:: python

            from petpal.pipelines.pca_guided_idif_steps import PCAGuidedTopVoxelsIDIFStep

            ## Initializing the step
            step = PCAGuidedTopVoxelsIDIFStep(input_image_path='/path/to/image.nii',
                                              mask_image_path='/path/to/mask.nii',
                                              output_array_path='/path/to/output.tsv',
                                              num_pca_components=3,
                                              verbose=False,
                                              selected_component=2,
                                              num_of_voxels=50)
            ## Running the step and saving the results
            step()

    """
    def __init__(self,
                 input_image_path: str,
                 mask_image_path: str,
                 output_array_path: str,
                 num_pca_components: int,
                 verbose: bool,
                 selected_component: int,
                 num_of_voxels: int):
        r"""Initializes the PCA-guided top voxels IDIF pipeline step.

        Args:
            input_image_path (str): Path to the input 4D-PET image.
            mask_image_path (str): Path to the mask image in the same space as 4D-PET image.
            output_array_path (str): Path where the resulting TAC data will be saved.
            num_pca_components (int): Number of PCA components to consider.
            verbose (bool): If `True`, enables detailed diagnostic output.
            selected_component (int): PCA component to use for voxel ranking.
            num_of_voxels (int): Number of top voxels to select.

        """
        PCAGuidedIDIFMixin.__init__(self,
                                    input_image_path=input_image_path,
                                    mask_image_path=mask_image_path,
                                    output_tac_path=output_array_path,
                                    num_pca_components=num_pca_components,
                                    verbose=verbose)
        ObjectBasedStep.__init__(self,
                                 name='pca_guided_top_voxels_idif',
                                 class_type=pca_guided_idif.PCAGuidedTopVoxelsIDIF,
                                 init_kwargs={'input_image_path'  : input_image_path,
                                              'mask_image_path'   : mask_image_path,
                                              'output_tac_path'   : output_array_path,
                                              'num_pca_components': num_pca_components,
                                              'verbose'           : verbose
                                              },
                                 call_kwargs=(self.call_kwargs | {'selected_component': selected_component,
                                                                  'num_of_voxels'     : num_of_voxels})
                                 )

        self.selected_component = selected_component
        self.num_of_voxels = num_of_voxels

    @property
    def selected_component(self):
        return self.call_kwargs['selected_component']

    @selected_component.setter
    def selected_component(self, value):
        self.call_kwargs['selected_component'] = value

    @property
    def num_of_voxels(self):
        return self.call_kwargs['num_of_voxels']

    @num_of_voxels.setter
    def num_of_voxels(self, value):
        self.call_kwargs['num_of_voxels'] = value

    @classmethod
    def default_pca_guided_idif_top_voxels(cls, name: str = 'pca_guided_top_voxels_idif', verbose=False, **overrides):
        r"""Creates a default instance for PCA-guided IDIF calculation based on selecting the top-ranked voxels from a single PCA component.


         All paths are set to empty strings by default.

        Args:
            name (str): Name of the step. Defaults to 'pca_guided_top_voxels_idif'.
            verbose (bool): Whether to enable verbose mode (default: False).
            **overrides: Parameter values to override the default settings.

        Returns:
            PCAGuidedTopVoxelsIDIFStep: A new instance of the PCA-guided top voxels IDIF step.

        Default Parameters:
            - input_image_path: '' (empty string)
            - mask_image_path: '' (empty string)
            - output_array_path: '' (empty string)
            - num_pca_components: 3
            - verbose: False
            - selected_component: 0
            - num_of_voxels: 50

        Side Effects:
            - If an invalid override is provided, a warning is issued, and the default instance is returned instead.

        """
        defaults = dict(input_image_path = '',
                        mask_image_path = '',
                        output_array_path = '',
                        num_pca_components = 3,
                        verbose = verbose,
                        selected_component = 0,
                        num_of_voxels = 50)

        overrides_dict = defaults | overrides

        try:
            out_class = cls(**overrides_dict)
        except TypeError as err:
            warnings.warn(f"Invalid override: {err}. Using default instance instead.", stacklevel=2)
            out_class = cls(**defaults)
        out_class.name = name
        return out_class



from . import graphical_analysis
from . import parametric_images
from . import reference_tissue_models
from . import tac_fitting
from . import tac_interpolation
from . import tcms_as_convolutions
from . import fit_tac_with_rtms
from . import rtm_analysis
from . import tac_uncertainty


def main():
    print("PETPAL - Kinetic Modeling")


if __name__ == "__main__":
    main()
    


"""
Module to handle abstracted functionalities
"""
from collections.abc import Callable
from pathlib import Path
import re
import os
import nibabel
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import ants

from . import image_io, math_lib, scan_timing

FULL_NAME = [
    'Background',
    'CorticalGrayMatter',
    'SubcorticalGrayMatter',
    'GrayMatter',
    'gm',
    'WhiteMatter',
    'wm',
    'CerebrospinalFluid',
    'Bone',
    'SoftTissue',
    'Nonbrain',
    'Lesion',
    'Brainstem',
    'Cerebellum'
]
SHORT_NAME = [
    'BG',
    'CGM',
    'SGM',
    'GM',
    'GM',
    'WM',
    'WM',
    'CSF',
    'B',
    'ST',
    'NB',
    'L',
    'BS',
    'CBM'
]


def make_path(paths: list[str]):
    """
    Creates a new path in local system by joining paths, and making any new directories, if
    necessary.

    Args:
        paths (list[str]): A list containing strings to be joined as a path in the system
            directory.

    Note:
        If the final string provided includes a period '.' (a proxy for checking if the path is a 
        file name) this method will result in creating the folder above the last provided string in
        the list.
    """
    end_dir = paths[-1]
    if end_dir.find('.') == -1:
        out_path = os.path.join(paths)
    else:
        out_path = os.path.join(paths[:-1])
    os.makedirs(out_path,exist_ok=True)


def abbreviate_region(region_name: str):
    """
    Converts long region names to their associated abbreviations.
    """
    name_out = region_name.replace('-','').replace('_','')
    for i,_d in enumerate(FULL_NAME):
        full_name = FULL_NAME[i]
        short_name = SHORT_NAME[i]
        name_out = name_out.replace(full_name,short_name)
    return name_out


def build_label_map(region_names: list[str]):
    """
    Builds a BIDS compliant label map. Loop through CTAB and convert names to
    abbreviations using :meth:`abbreviate_region`
    """
    abbreviated_names = list(map(abbreviate_region,region_names))
    return abbreviated_names


def weighted_series_sum(input_image_path: str,
                        out_image_path: str,
                        half_life: float,
                        verbose: bool=False,
                        start_time: float=0,
                        end_time: float=-1) -> np.ndarray:
    r"""
    Sum a 4D image series weighted based on time and re-corrected for decay correction.

    First, a scaled image is produced by multiplying each frame by its length in seconds,
    and dividing by the decay correction applied:

    .. math::

        f_i'=f_i\times \frac{t_i}{d_i}

    Where :math:`f_i,t_i,d_i` are the i-th frame, frame duration, and decay correction factor of
    the PET series. This scaled image is summed over the time axis. Then, to get the output, we
    multiply by a factor called ``total decay`` and divide by the full length of the image:

    .. math::

        d_{S} = \frac{\lambda*t_{S}}{(1-\exp(-\lambda*t_{S}))(\exp(\lambda*t_{0}))}

    .. math::

        S(f) = \sum(f_i') * d_{S} / t_{S}

    where :math:`\lambda=\log(2)/T_{1/2}` is the decay constant of the radio isotope,
    :math:`t_0` is the start time of the first frame in the PET series, the subscript :math:`S`
    indicates the total quantity computed over all frames, and :math:`S(f)` is the final weighted
    sum image.

    # TODO: Determine half_life from .json rather than passing as argument.

    Args:
        input_image_path (str): Path to a .nii or .nii.gz file containing a 4D
            PET image on which the weighted sum is calculated. Assume a metadata
            file exists with the same path and file name, but with extension .json,
            and follows BIDS standard.
        out_image_path (str): Path to a .nii or .nii.gz file to which the weighted
            sum is written. If none, will not write output to a file.
        half_life (float): Half life of the PET radioisotope in seconds.
        verbose (bool): Set to ``True`` to output processing information. Default is False.
        start_time (float): Time, relative to scan start in seconds, at which
            calculation begins. Must be used with ``end_time``. Default value 0.
        end_time (float): Time, relative to scan start in seconds, at which
            calculation ends. Use value ``-1`` to use all frames in image series.
            If equal to ``start_time``, one frame at start_time is used. Default value -1.

    Returns:
        np.ndarray: 3D image array, in the same space as the input, with the weighted sum
            calculation applied.

    Raises:
        ValueError: If ``half_life`` is zero or negative.
    """
    if half_life <= 0:
        raise ValueError('(ImageOps4d): Radioisotope half life is zero or negative.')
    pet_meta = image_io.load_metadata_for_nifti_with_same_filename(input_image_path)
    pet_image = nibabel.load(input_image_path)
    pet_series = pet_image.get_fdata()
    frame_start = pet_meta['FrameTimesStart']
    frame_duration = pet_meta['FrameDuration']

    if 'DecayCorrectionFactor' in pet_meta.keys():
        decay_correction = pet_meta['DecayCorrectionFactor']
    elif 'DecayFactor' in pet_meta.keys():
        decay_correction = pet_meta['DecayFactor']
    else:
        raise ValueError("Neither 'DecayCorrectionFactor' nor 'DecayFactor' exist in meta-data "
                         "file")

    if 'TracerRadionuclide' in pet_meta.keys():
        tracer_isotope = pet_meta['TracerRadionuclide']
        if verbose:
            print(f"(ImageOps4d): Radio isotope is {tracer_isotope} "
                f"with half life {half_life} s")

    if end_time==-1:
        pet_series_adjusted = pet_series
        frame_start_adjusted = frame_start
        frame_duration_adjusted = frame_duration
        decay_correction_adjusted = decay_correction
    else:
        scan_start = frame_start[0]
        nearest_frame = interp1d(x=frame_start,
                                 y=range(len(frame_start)),
                                 kind='nearest',
                                 bounds_error=False,
                                 fill_value='extrapolate')
        calc_first_frame = int(nearest_frame(start_time+scan_start))
        calc_last_frame = int(nearest_frame(end_time+scan_start))
        if calc_first_frame==calc_last_frame:
            calc_last_frame += 1
        pet_series_adjusted = pet_series[:,:,:,calc_first_frame:calc_last_frame]
        frame_start_adjusted = frame_start[calc_first_frame:calc_last_frame]
        frame_duration_adjusted = frame_duration[calc_first_frame:calc_last_frame]
        decay_correction_adjusted = decay_correction[calc_first_frame:calc_last_frame]

    wsc = math_lib.weighted_sum_computation
    image_weighted_sum = wsc(frame_duration=frame_duration_adjusted,
                             half_life=half_life,
                             pet_series=pet_series_adjusted,
                             frame_start=frame_start_adjusted,
                             decay_correction=decay_correction_adjusted)

    if out_image_path is not None:
        pet_sum_image = nibabel.nifti1.Nifti1Image(dataobj=image_weighted_sum,
                                                   affine=pet_image.affine,
                                                   header=pet_image.header)
        nibabel.save(pet_sum_image, out_image_path)
        if verbose:
            print(f"(ImageOps4d): weighted sum image saved to {out_image_path}")
        image_io.safe_copy_meta(input_image_path=input_image_path,
                                out_image_path=out_image_path)

    return image_weighted_sum

def weighted_series_sum_over_window_indices(input_image_4d: ants.core.ANTsImage | str,
                                            output_image_path: str | None,
                                            window_start_id: int,
                                            window_end_id: int,
                                            half_life: float,
                                            image_frame_info: scan_timing.ScanTimingInfo) -> ants.core.ANTsImage | None:
    r"""
    Computes a weighted series sum over a specified window of indices for a 4D PET image.

    Args:
        input_image_4d (ants.core.ANTsImage | str): Input 4D PET image as an ANTs image or the path to a NIfTI file.
        output_image_path (str | None): Path to save the output image. If `None`, the output is not saved.
        window_start_id (int): Start index of the image window.
        window_end_id (int): End index of the image window.
        half_life (float): Radioactive tracer's half-life in seconds.
        image_frame_info (scan_timing.ScanTimingInfo): Frame timing information with:
            - duration (np.ndarray): Frame durations.
            - start (np.ndarray): Frame start times.
            - end (np.ndarray): Frame ends.
            - center (np.ndarray): Frame centers.
            - decay (np.ndarray): Decay correction factors.

    Returns:
        ants.core.ANTsImage: Resultant image after weighted sum computation.

    Note:
        If `output_image_path` is provided, the computed image will be saved to the specified path.
        This allows us to utilize ANTs pipelines
        ``weighted_series_sum_over_window_indices(...).get_center_of_mass()`` for example.

    """
    if isinstance(input_image_4d, str):
        input_image_4d = image_io.safe_load_4dpet_nifti(filename=input_image_4d)

    assert len(input_image_4d.shape) == 4, "Input image must be 4D."

    window_wss = math_lib.weighted_sum_computation_over_index_window(pet_series=input_image_4d,
                                                                     window_start_id=window_start_id,
                                                                     window_end_id=window_end_id,
                                                                     half_life=half_life,
                                                                     frame_starts=image_frame_info.start,
                                                                     frame_durations=image_frame_info.duration,
                                                                     decay_factors=image_frame_info.decay,)

    window_wss = ants.from_numpy(data=window_wss,
                                 origin=input_image_4d.origin[:-1],
                                 direction=input_image_4d.direction[:-1],
                                 spacing=input_image_4d.spacing[:-1])

    if output_image_path is not None:
        ants.image_write(image=window_wss, filename=output_image_path)

    return  window_wss


def read_plasma_glucose_concentration(file_path: str, correction_scale: float = 1.0 / 18.0) -> float:
    r"""
    Temporary hacky function to read a single plasma glucose concentration value from a file.

    This function reads a single numerical value from a specified file and applies a correction scale to it.
    The primary use is to quickly extract plasma glucose concentration for further processing. The default
    scaling os 1.0/18.0 is the one used in the CMMS study to get the right units.

    Args:
        file_path (str): Path to the file containing the plasma glucose concentration value.
        correction_scale (float): Scale factor for correcting the read value. Default is `1.0/18.0`.

    Returns:
        float: Corrected plasma glucose concentration value.
    """
    return correction_scale * float(np.loadtxt(file_path))


def convert_ctab_to_dseg(ctab_path: str,
                         dseg_path: str,
                         column_names: list[str]=None):
    """
    Convert a FreeSurfer compatible color table into a BIDS compatible label
    map ``dseg.tsv``.

    Args:
        ctab_path (str): Path to FreeSurfer compatible color table.
        dseg_path (str): Path to ``dseg.tsv`` label mapfile to save.
        column_names (list[str]): List of columns present in color table. Must
            include 'mapping' and 'name'.
    """
    if column_names==None:
        column_names = ['mapping','name','r','g','b','a','ttype']
    fs_ctab = pd.read_csv(ctab_path,
                          delim_whitespace=True,
                          header=None,
                          comment='#',
                          names=column_names)
    label_names = {'name': fs_ctab['name'],
                   'mapping': fs_ctab['mapping'],
                   'abbreviation': build_label_map(fs_ctab['name'])}
    label_map = pd.DataFrame(data=label_names,
                             columns=['name','abbreviation','mapping']).rename_axis('index')
    label_map = label_map.sort_values(by=['mapping'])
    label_map.to_csv(dseg_path,sep='\t')
    return label_map


def capitalize_first_char_of_str(input_str: str) -> str:
    """
    Capitalize only the first character of a string, leaving the remainder unchanged.
    Args:
        input_str (str): The string to capitalize the first character of.
    Returns:
        output_str (str): The string with only the first character capitalized.
    """
    output_str = input_str[0].capitalize()+input_str[1:]
    return output_str


def str_to_camel_case(input_str) -> str:
    """
    Take a string and return the string converted to camel case.

    Special characters (? * - _ / \\) are removed and treated as word separaters. Different
    words are then capitalized at the first character, leaving other alphanumeric characters
    unchanged.

    Args:
        input_str (str): The string to convert to camel case and remove special characters.
    Returns:
        camel_case_str (str): The string converted to camel case (e.g. CamelCase) with special
            characters removed.
    """
    split_str = re.split(r'[-_?*/\\]', input_str)
    capped_split_str = []
    capitalize_first = capitalize_first_char_of_str
    for part in split_str:
        capped_str = capitalize_first(input_str=part)
        capped_split_str += [capped_str]
    camel_case_str = ''.join(capped_split_str)
    return camel_case_str


def nearest_frame_to_timepoint(frame_times: np.ndarray) -> Callable[[float],float]:
    """Returns a step function that gets the index of the frame closest to a provided timepoint
    based on an array of frame times, such as the frame starts or reference times.

    Args:
        frame_times (np.ndarray): The frame times on which to generate the step function.

    Returns:
        nearest_frame_func (Callable[[float],float]): A function that returns the time closest to
            the provided timepoint.
    """
    nearest_frame_func = interp1d(x=frame_times,
                                  y=range(len(frame_times)),
                                  kind='nearest',
                                  bounds_error=False,
                                  fill_value='extrapolate')
    return nearest_frame_func


def get_average_of_timeseries(input_image: ants.ANTsImage) -> ants.ANTsImage:
    """
    Get average of a 4D ANTsImage and return as a 3D ANTsImage.

    Args:
        input_image (ants.ANTsImage): 4D PET image over which to compute timeseries average.

    Returns:
        mean_image (ants.ANTsImage): 3D mean over time in the PET image.
    """
    assert len(input_image.shape) == 4, "Input image must be 4D"
    mean_array = input_image.mean(axis=-1)
    mean_image = ants.from_numpy(data=mean_array,
                                 origin=input_image.origin[:-1],
                                 spacing=input_image.spacing[:-1],
                                 direction=input_image.direction[:-1,:-1])
    return mean_image


def coerce_outpath_extension(path: str, ext: str) -> str:
    """Coerce a path to the same absolute path with a provided filetype extension.
    
    Args:
        path (str): Path to a file.
        ext (str): Desired output extension.
    
    Returns:
        abs_path_with_extension (str): Absolute path of the input file with the modified extension.
    
    Example:

        .. code-block:: python

            from petpal.utils.useful_functions import coerce_outpath_extension
            
            my_path = 'my_file.nii.gz'
            my_csv_file = coerce_outpath_extension(my_path, '.csv')
            print(my_csv_file) # prints '/current/working/directory/my_file.csv'

    """
    path_obj = Path(path)
    while path_obj.suffix!='':
        path_obj = path_obj.with_suffix('')
    path_obj_with_suffix = path_obj.with_suffix(ext)
    return str(path_obj_with_suffix.absolute())



r"""
Todo:
    * Method to calculate chi-squared from fit, 
        :math:`\chi^2=\Sigma_{i=1}^{n}(x_i-x_{ti})^2/\sigma_i^2`
    
"""
import numpy as np
from scipy.optimize import curve_fit as sp_fit
import numba
from .graphical_analysis import get_index_from_threshold
from .graphical_analysis import cumulative_trapezoidal_integral as cum_trapz
from . import tcms_as_convolutions as tcms_conv

def calc_srtm_tac(tac_times_in_minutes: np.ndarray,
                  ref_tac_vals: np.ndarray, r1: float, k2: float, bp: float) -> np.ndarray:
    r"""
    Calculate the Time Activity Curve (TAC) using the Simplified Reference Tissue Model (SRTM)
    with the given reference TAC and kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.


    The SRTM TAC can be calculated as:

    .. math::

        C(t)=R_{1}C_\mathrm{R}(t) + \left(k_{2} - \frac{R_{1}k_{2}}{1+\mathrm{BP}}\right)
        C_\mathrm{R}(t)\otimes\exp\left(- \frac{k_{2}t}{1+\mathrm{BP}}\right),


    where :math:`C_\mathrm{R}(t)` is the reference TAC, :math:`R_{1}=\frac{k_1^\prime}{k_1}`,
    :math:`k_{2}` is the rate-constant from the tissue compartment to plasma, and
    :math:`\mathrm{BP}` is the binding potential.


    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        r1 (float): The ratio of the clearance rate of tracer from plasma to the reference to the
             transfer rate of the tracer from plasma to the tissue;
             :math:`R_{1}\equiv\frac{k_1^\prime}{k_1}`.
        k2 (float): The rate constant for the transfer of the tracer from tissue compartment to
            plasma.
        bp (float): The binding potential of the tracer in the tissue.
        ref_tac_vals (np.ndarray): The values of the reference TAC.

    Returns:
        np.ndarray: TAC values calculated using SRTM.


    Raises:
        AssertionError: If the reference TAC and times are different dimensions.

    """
    first_term = r1 * ref_tac_vals
    bp_coeff = k2 / (1.0 + bp)
    exp_term = np.exp(-bp_coeff * tac_times_in_minutes)
    dt = tac_times_in_minutes[1] - tac_times_in_minutes[0]
    convolution_term = tcms_conv.calc_convolution_with_check(f=exp_term,g=ref_tac_vals,dt=dt)
    second_term = (k2 - r1 * bp_coeff) * convolution_term

    return first_term + second_term


def _calc_simplified_frtm_tac(tac_times_in_minutes: np.ndarray,
                              ref_tac_vals: np.ndarray,
                              r1: float,
                              a1: float,
                              a2: float,
                              alpha_1: float,
                              alpha_2: float) -> np.ndarray:
    r"""
    Calculate the Time Activity Curve (TAC) for the Full Reference Tissue Model (FRTM) given the
    reference TAC and simplified coefficients. The coefficients can be generated from kinetic
    constants using :func:`_calc_frtm_params_from_kinetic_params`

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    We use a more compact form for the FRTM:

    .. math::

        C(t) = R_{1}C_\mathrm{R}(t) + \left[ A_{1}e^{-\alpha_{1}t} + A_{2}e^{-\alpha_{2}t} \right]
        \otimes C_\mathrm{R}(t),

    where :math:`R_{1}\equiv\frac{k_1^\prime}{k_1}`, and
    :math:`A_{1},\,A_{2},\,\alpha_1,\,\alpha_2` can be calculated from the underlying kinetic
    constants. See :func:`_calc_frtm_params_from_kinetic_params` for more details about
    the parameter calculation.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        r1 (float): The ratio of the clearance rate of tracer from plasma to the reference to the transfer rate of the
            tracer from plasma to the tissue; :math:`R_{1}\equiv\frac{k_1^\prime}{k_1}`.
        a1 (float): Coefficient of the first exponential term.
        a2 (float): Coefficient of the second exponential term.
        alpha_1 (float): Coefficient inside the first exponential.
        alpha_2 (float): Coefficient inside the second exponential.
        ref_tac_vals (np.ndarray): The values of the reference TAC.

    Returns:
        np.ndarray: TAC values calculated using FRTM.


    Raises:
        AssertionError: If the reference TAC and times are different dimensions.

    """
    first_term = r1 * ref_tac_vals
    exp_funcs = a1 * np.exp(-alpha_1 * tac_times_in_minutes) + a2 * np.exp(-alpha_2 * tac_times_in_minutes)
    dt = tac_times_in_minutes[1] - tac_times_in_minutes[0]
    second_term = tcms_conv.calc_convolution_with_check(f=exp_funcs, g=ref_tac_vals, dt=dt)
    return first_term + second_term


def _calc_frtm_params_from_kinetic_params(r1: float,
                                          k2: float,
                                          k3: float,
                                          k4: float) -> tuple[float, float, float, float, float]:
    r"""
    Calculates the parameters (coefficients) for the simplified FRTM function
    (:func:`_calc_simplified_frtm_tac`) given the kinetic constants.

    The parameters are defined as:

    .. math::

        \alpha_{1} &= \frac{\beta - \chi}{2}\\
        \alpha_{2} &= \frac{\beta + \chi}{2}\\
        A_{1} &= \left(\frac{k_{3} + k_{4} -\alpha_{2}}{\chi} \right)\left( \frac{k_{2}}{R_{1}} -
        \alpha_{2} \right)\\
        A_{2} &= \left(\frac{\alpha_{1}-k_{3} - k_{4} }{\chi} \right)\left( \frac{k_{2}}{R_{1}} -
        \alpha_{1} \right),

    where additionally we have:

    .. math::

        \alpha_{1} &= \frac{k_{2} + k_{3} + k_{4} - \sqrt{\left( k_{2} + k_{3} + k_{4} \right)^2 -
        4k_{2}k_{4}}}{2}\\
        \alpha_{2} &= \frac{k_{2} + k_{3} + k_{4} + \sqrt{\left( k_{2} + k_{3} + k_{4} \right)^2 -
        4k_{2}k_{4}}}{2}\\
        A_{1} &= \left( \frac{k_{3} + k_{4} -\alpha_{2}}{\alpha_{1} - \alpha_{2}} \right)\left(
        \frac{k_{2}}{R_{1}}
        - \alpha_{2} \right)\\
        A_{2} &= \left(  \frac{\alpha_{1}-k_{3} - k_{4} }{\alpha_{1} - \alpha_{2}} \right)\left(
        \frac{k_{2}}{R_{1}}
        - \alpha_{1} \right)


    Args:
        r1 (float): The ratio of the clearance rate of tracer from plasma to the reference to the
            transfer rate of the tracer from plasma to the tissue;
            :math:`R_{1}\equiv\frac{k_1^\prime}{k_1}`.
        k2 (float): The rate of tracer transfer from the first tissue compartment to plasma.
        k3 (float): The rate of tracer transfer from the first tissue compartment to the second
            tissue compartment.
        k4 (float): The rate of tracer transfer from the second tissue compartment to the first
            tissue compartment.

    Returns:
        tuple: (``r1``, ``a1``, ``a2``, ``alpha_1``, ``alpha_2``) parameters for
            :func:`_calc_simplified_frtm_tac`.
    """
    beta = k2 + k3 + k4
    chi = np.sqrt(beta ** 2. - 4.0 * k2 * k4)
    alpha_1 = (beta - chi) / 2.0
    alpha_2 = (beta + chi) / 2.0
    a1 = (k3 + k4 - alpha_2) / chi * (k2 / r1 - alpha_2)
    a2 = (alpha_1 - k3 - k4) / chi * (k2 / r1 - alpha_1)
    return r1, a1, a2, alpha_1, alpha_2


def calc_frtm_tac(tac_times_in_minutes: np.ndarray,
                  ref_tac_vals: np.ndarray,
                  r1: float,
                  k2: float,
                  k3: float,
                  k4: float) -> np.ndarray:
    r"""
    Calculate the Time Activity Curve (TAC) using the Full Reference Tissue Model (SRTM) with the
    given reference TAC and kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    The FRTM TAC can be calculated as:

    .. math::

        C(t) = R_{1}C_\mathrm{R}(t) + \left[ A_{1}e^{-\alpha_{1}t} + A_{2}e^{-\alpha_{2}t} \right]
        \otimes C_\mathrm{R}(t),

    where additionally we have:

    .. math::

        \alpha_{1} &= \frac{k_{2} + k_{3} + k_{4} - \sqrt{\left( k_{2} + k_{3} + k_{4} \right)^2 -
        4k_{2}k_{4}}}{2}\\
        \alpha_{2} &= \frac{k_{2} + k_{3} + k_{4} + \sqrt{\left( k_{2} + k_{3} + k_{4} \right)^2 -
        4k_{2}k_{4}}}{2}\\
        A_{1} &= \left( \frac{k_{3} + k_{4} -\alpha_{2}}{\alpha_{1} - \alpha_{2}} \right)\left(
        \frac{k_{2}}{R_{1}}
        - \alpha_{2} \right)\\
        A_{2} &= \left(  \frac{\alpha_{1}-k_{3} - k_{4} }{\alpha_{1} - \alpha_{2}} \right)\left(
        \frac{k_{2}}{R_{1}}
        - \alpha_{1} \right)


    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        r1 (float): The ratio of the clearance rate of tracer from plasma to the reference to the transfer rate of the
            tracer from plasma to the tissue; :math:`R_{1}\equiv\frac{k_1^\prime}{k_1}`.
        k2 (float): The rate of tracer transfer from the first tissue compartment to plasma.
        k3 (float): The rate of tracer transfer from the first tissue compartment to the second
            tissue compartment.
        k4 (float): The rate of tracer transfer from the second tissue compartment to the first
            tissue compartment.
        ref_tac_vals (np.ndarray): The values of the reference TAC.

    Returns:
        np.ndarray: TAC values calculated using FRTM.

    Raises:
        AssertionError: If the reference TAC and times are different dimensions.

    See Also:
        * :func:`_calc_simplified_frtm`
        * :func:`_calc_frtm_params_from_kinetic_params`

    """
    r1_n, a1, a2, alpha_1, alpha_2 = _calc_frtm_params_from_kinetic_params(r1=r1, k2=k2, k3=k3, k4=k4)
    return _calc_simplified_frtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1_n, a1=a1, a2=a2,
                                     alpha_1=alpha_1, alpha_2=alpha_2)


def fit_srtm_to_tac(tac_times_in_minutes: np.ndarray,
                    tgt_tac_vals: np.ndarray,
                    ref_tac_vals: np.ndarray,
                    r1_start: float = 0.5,
                    k2_start: float = 0.5,
                    bp_start: float = 0.5) -> tuple:
    r"""
    Fit SRTM to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    starting guesses for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This is a simple wrapper around :func:`scipy.optimize.curve_fit` and does not use any bounds
    for the different parameters.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM.
        ref_tac_vals (np.ndarray): Reference (and Target) TAC times.
        r1_start (float): Starting guess for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}` parameter.
        k2_start (float): Starting guess for :math:`k_2` parameter.
        bp_start (float): Starting guess for the binding potential.

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from
            :func:`scipy.optimize.curve_fit`

    Raises:
        AssertionError: If the reference TAC and times are different dimensions.

    See Also:
        * :func:`calc_srtm_tac`

    """
    def _fitting_srtm(tac_times_in_minutes, r1, k2, bp):
        return calc_srtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1, k2=k2, bp=bp)
    
    starting_values = [r1_start, k2_start, bp_start]
    
    return sp_fit(f=_fitting_srtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals, p0=starting_values)


def fit_srtm2_to_tac(tac_times_in_minutes: np.ndarray,
                     tgt_tac_vals: np.ndarray,
                     ref_tac_vals: np.ndarray,
                     k2_prime: float = 0.5,
                     r1_start: float = 0.5,
                     bp_start: float = 0.5) -> tuple:
    r"""
    Fit SRTM2 to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    starting guesses for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This is a simple wrapper around :func:`scipy.optimize.curve_fit` and does not use any bounds
    for the different parameters.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM2.
        ref_tac_vals (np.ndarray): Reference (and Target) TAC times.
        k2_prime (float): The :math:`k_2^\prime` value.` Defaults to 0.5.
        r1_start (float): Starting guess for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}` parameter.
        bp_start (float): Starting guess for the binding potential.

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from
            :func:`scipy.optimize.curve_fit`

    Raises:
        AssertionError: If the reference TAC and times are different dimensions.

    See Also:
        * :func:`calc_srtm_tac`
        * :func:`fit_srtm_to_tac`

    """

    def _fitting_srtm(tac_times_in_minutes, r1, bp):
        return calc_srtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1, k2=k2_prime, bp=bp)

    starting_values = [r1_start, bp_start]

    return sp_fit(f=_fitting_srtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals, p0=starting_values)



def fit_srtm_to_tac_with_bounds(tac_times_in_minutes: np.ndarray,
                                tgt_tac_vals: np.ndarray,
                                ref_tac_vals: np.ndarray,
                                r1_bounds: np.ndarray = np.asarray(
                                    [0.5, 0.0, 10.0]),
                                k2_bounds: np.ndarray = np.asarray(
                                    [0.5, 0.0, 10.0]),
                                bp_bounds: np.ndarray = np.asarray([0.5, 0.0, 10.0])) -> tuple:
    r"""
    Fit SRTM to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    bounds for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This function is a wrapper around `scipy.optimize.curve_fit` and uses parameter bounds during
    optimization. The bounds for each parameter are formatted as:
    ``(starting_value, lo_bound, hi_bound)``.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM.
        ref_tac_vals (np.ndarray): Reference TAC values.
        r1_bounds (np.ndarray): The bounds for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}` parameter.
            Defaults to [0.5, 0.0, 10.0].
        k2_bounds (np.ndarray): The bounds for :math:`k_2` parameter. Defaults to [0.5, 0.0, 10.0].
        bp_bounds (np.ndarray): The bounds for the binding potential parameter. Defaults to
            [0.5, 0.0, 10.0].

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from `scipy.optimize.curve_fit`.

    Raises:
        AssertionError: If the target TAC and times are different dimensions.

    See Also:
        * :func:`calc_srtm_tac`

    """
    def _fitting_srtm(tac_times_in_minutes, r1, k2, bp):
        return calc_srtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1, k2=k2, bp=bp)

    st_values = (r1_bounds[0], k2_bounds[0], bp_bounds[0])
    lo_values = (r1_bounds[1], k2_bounds[1], bp_bounds[1])
    hi_values = (r1_bounds[2], k2_bounds[2], bp_bounds[2])

    return sp_fit(f=_fitting_srtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals,
                  p0=st_values, bounds=[lo_values, hi_values])


def fit_srtm2_to_tac_with_bounds(tac_times_in_minutes: np.ndarray,
                                 tgt_tac_vals: np.ndarray,
                                 ref_tac_vals: np.ndarray,
                                 k2_prime: float = 0.5,
                                 r1_bounds: np.ndarray = np.asarray(
                                     [0.5, 0.0, 10.0]),
                                 bp_bounds: np.ndarray = np.asarray([0.5, 0.0, 10.0])) -> tuple:
    r"""
    Fit SRTM2 to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    bounds for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This function is a wrapper around `scipy.optimize.curve_fit` and uses parameter bounds during
    optimization. The bounds for each parameter are formatted as:
    ``(starting_value, lo_bound, hi_bound)``.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM2.
        ref_tac_vals (np.ndarray): Reference TAC values.
        k2_prime (int): The value for :math:`k_2^\prime`. Defaults to 0.5.
        r1_bounds (np.ndarray): The bounds for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}`
            parameter. Defaults to [0.5, 0.0, 10.0].
        bp_bounds (np.ndarray): The bounds for the binding potential parameter.
            Defaults to [0.5, 0.0, 10.0].

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from `scipy.optimize.curve_fit`.

    Raises:
        AssertionError: If the target TAC and times are different dimensions.

    See Also:
        * :func:`calc_srtm_tac`
        * :func:`fit_srtm2_tac`
        * :func:`fit_srtm_to_tac_with_bounds`

    """

    def _fitting_srtm(tac_times_in_minutes, r1, bp):
        return calc_srtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1, k2=k2_prime, bp=bp)

    st_values = (r1_bounds[0], bp_bounds[0])
    lo_values = (r1_bounds[1], bp_bounds[1])
    hi_values = (r1_bounds[2], bp_bounds[2])

    return sp_fit(f=_fitting_srtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals, p0=st_values, bounds=[lo_values, hi_values])


def fit_frtm_to_tac(tac_times_in_minutes: np.ndarray,
                    tgt_tac_vals: np.ndarray,
                    ref_tac_vals: np.ndarray,
                    r1_start: float = 0.5,
                    k2_start: float = 0.5,
                    k3_start: float = 0.5,
                    k4_start: float = 0.5) -> tuple:
    r"""
    Fit FRTM to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    starting guesses for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This is a simple wrapper around :func:`scipy.optimize.curve_fit` and does not use any bounds
    for the different parameters.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM.
        ref_tac_vals (np.ndarray): Reference (and Target) TAC times.
        r1_start (float): Starting guess for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}` parameter.
        k2_start (float): Starting guess for :math:`k_2` parameter.
        k3_start (float): Starting guess for :math:`k_3` parameter.
        k4_start (float): Starting guess for :math:`k_4` parameter.

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from
            :func:`scipy.optimize.curve_fit`

    Raises:
        AssertionError: If the reference TAC and times are different dimensions.

    See Also:
        * :func:`calc_frtm_tac`

    """
    def _fitting_frtm(tac_times_in_minutes, r1, k2, k3, k4):
        return calc_frtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1, k2=k2, k3=k3, k4=k4)

    starting_values = (r1_start, k2_start, k3_start, k4_start)
    return sp_fit(f=_fitting_frtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals, p0=starting_values)


def fit_frtm2_to_tac(tac_times_in_minutes: np.ndarray,
                     tgt_tac_vals: np.ndarray,
                     ref_tac_vals: np.ndarray,
                     k2_prime: float = 0.5,
                     r1_start: float = 0.5,
                     k3_start: float = 0.5,
                     k4_start: float = 0.5) -> tuple:
    r"""
    Fit FRTM2 to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    starting guesses for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This is a simple wrapper around :func:`scipy.optimize.curve_fit` and does not use any bounds
    for the different parameters.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM.
        ref_tac_vals (np.ndarray): Reference (and Target) TAC times.
        k2_prime (float): Value for the :math:`k_2^\prime` parameter. Defaults to 0.5.
        r1_start (float): Starting guess for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}` parameter.
        k3_start (float): Starting guess for :math:`k_3` parameter.
        k4_start (float): Starting guess for :math:`k_4` parameter.

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from
            :func:`scipy.optimize.curve_fit`

    Raises:
        AssertionError: If the reference TAC and times are different dimensions.

    See Also:
        * :func:`calc_frtm_tac`

    """

    def _fitting_frtm(tac_times_in_minutes, r1, k3, k4):
        return calc_frtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1, k2=k2_prime, k3=k3, k4=k4)

    starting_values = (r1_start, k3_start, k4_start)
    return sp_fit(f=_fitting_frtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals, p0=starting_values)


def fit_frtm_to_tac_with_bounds(tac_times_in_minutes: np.ndarray,
                                tgt_tac_vals: np.ndarray,
                                ref_tac_vals: np.ndarray,
                                r1_bounds: np.ndarray = np.asarray(
                                    [0.5, 0.0, 10.0]),
                                k2_bounds: np.ndarray = np.asarray(
                                    [0.5, 0.0, 10.0]),
                                k3_bounds: np.ndarray = np.asarray(
                                    [0.5, 0.0, 10.0]),
                                k4_bounds: np.ndarray = np.asarray(
                                    [0.5, 0.0, 10.0])) -> tuple:
    r"""
    Fit FRTM to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    bounds for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This function is a wrapper around `scipy.optimize.curve_fit` and uses parameter bounds during
    optimization. The bounds for each parameter are formatted as:
    ``(starting_value, lo_bound, hi_bound)``.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM.
        ref_tac_vals (np.ndarray): Reference TAC values.
        r1_bounds (np.ndarray): The bounds for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}`
            parameter. Defaults to [0.5, 0.0, 10.0].
        k2_bounds (np.ndarray): The bounds for :math:`k_2` parameter. Defaults to [0.5, 0.0, 10.0].
        k3_bounds (np.ndarray): The bounds for :math:`k_3` parameter. Defaults to [0.5, 0.0, 10.0].
        k4_bounds (np.ndarray): The bounds for :math:`k_4` parameter. Defaults to [0.5, 0.0, 10.0].

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from `scipy.optimize.curve_fit`.

    Raises:
        AssertionError: If the target TAC and times are different dimensions.

    See Also:
        * :func:`calc_frtm_tac`

    """
    def _fitting_frtm(tac_times_in_minutes, r1, k2, k3, k4):
        return calc_frtm_tac(tac_times_in_minutes=tac_times_in_minutes,
                             ref_tac_vals=ref_tac_vals, r1=r1, k2=k2, k3=k3, k4=k4)

    st_values = (r1_bounds[0], k2_bounds[0], k3_bounds[0], k4_bounds[0])
    lo_values = (r1_bounds[1], k2_bounds[1], k3_bounds[1], k4_bounds[1])
    hi_values = (r1_bounds[2], k2_bounds[2], k3_bounds[2], k4_bounds[2])

    return sp_fit(f=_fitting_frtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals,
                  p0=st_values, bounds=[lo_values, hi_values])


def fit_frtm2_to_tac_with_bounds(tac_times_in_minutes: np.ndarray,
                                 tgt_tac_vals: np.ndarray,
                                 ref_tac_vals: np.ndarray,
                                 k2_prime: float = 0.5,
                                 r1_bounds: np.ndarray = np.asarray(
                                     [0.5, 0.0, 10.0]),
                                 k3_bounds: np.ndarray = np.asarray(
                                     [0.5, 0.0, 10.0]),
                                 k4_bounds: np.ndarray = np.asarray([0.5, 0.0, 10.0])) -> tuple:
    r"""
    Fit FRTM2 to the provided target Time Activity Curve (TAC), given the reference TAC, times, and
    bounds for the kinetic parameters.

    .. important::
        This function assumes that the reference TAC is uniformly sampled with respect to time
        since we perform convolutions.

    This function is a wrapper around `scipy.optimize.curve_fit` and uses parameter bounds during
    optimization. The bounds for each parameter are formatted as:
    ``(starting_value, lo_bound, hi_bound)``.

    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC to fit with the SRTM.
        ref_tac_vals (np.ndarray): Reference TAC values.
        k2_prime (float): The value for the :math:`k_2^\prime` parameter. Defaults to 0.5.
        r1_bounds (np.ndarray): The bounds for the :math:`R_1\equiv\frac{k_1^\prime}{k_1}`
            parameter. Defaults to [0.5, 0.0, 10.0].
        k3_bounds (np.ndarray): The bounds for :math:`k_3` parameter. Defaults to [0.5, 0.0, 10.0].
        k4_bounds (np.ndarray): The bounds for :math:`k_4` parameter. Defaults to [0.5, 0.0, 10.0].

    Returns:
        tuple: (``fit_parameters``, ``fit_covariance``). Output from `scipy.optimize.curve_fit`.

    Raises:
        AssertionError: If the target TAC and times are different dimensions.

    See Also:
        * :func:`calc_frtm_tac`
        * :func:`fit_frtm2_to_tac`

    """

    def _fitting_frtm(tac_times_in_minutes, r1, k3, k4):
        return calc_frtm_tac(tac_times_in_minutes=tac_times_in_minutes, ref_tac_vals=ref_tac_vals, r1=r1, k2=k2_prime, k3=k3, k4=k4)

    st_values = (r1_bounds[0], k3_bounds[0], k4_bounds[0])
    lo_values = (r1_bounds[1], k3_bounds[1], k4_bounds[1])
    hi_values = (r1_bounds[2], k3_bounds[2], k4_bounds[2])

    return sp_fit(f=_fitting_frtm, xdata=tac_times_in_minutes, ydata=tgt_tac_vals, p0=st_values, bounds=[lo_values, hi_values])

@numba.njit(fastmath=True)
def fit_mrtm_original_to_tac(tac_times_in_minutes: np.ndarray,
                             tgt_tac_vals: np.ndarray,
                             ref_tac_vals: np.ndarray,
                             t_thresh_in_mins: float,
                             weights: np.ndarray=None):
    r"""
    Fit the original (1996) Multilinear Reference Tissue Model (MRTM) to the provided target Time
    Activity Curve (TAC) values given the reference TAC, times, and threshold time (in minutes).
    The data are fit for all values beyond the threshold. We assume that the target TAC and
    reference TAC are sampled at the same times.

    .. important::
        This function assumes that both TACs are sampled at the same time, and that the time is in
        minutes.


    We have the following multilinear regression:

    .. math::

        \frac{\int_{0}^{T}C(t)\mathrm{d}t}{C(T)}=\frac{V}{V^{\prime}}
        \frac{\int_{0}^{T}C^{\prime}(t)\mathrm{d}t}{C(T)}
        - \frac{V}{V^{\prime}k_{2}^{\prime}} \frac{C^{\prime}(T)}{C(T)} + b


    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC values to fit the MRTM.
        ref_tac_vals (np.ndarray): Reference TAC values.
        t_thresh_in_mins (float): Threshold time in minutes.
        weights (np.ndarray): Weights applied to each frame. Default None.

    Returns:
        np.ndarray: Array containing fit results. (:math:`\frac{V}{V^{\prime}}`,
        :math:`\frac{V}{V^{\prime}k_{2}^{\prime}}`, :math:`b`)

    Note:
        This function is implemented with numba for improved performance.

    """
    if weights is None:
        weights = np.ones_like(tac_times_in_minutes)

    non_zero_indices = np.argwhere(tgt_tac_vals != 0.).T[0]

    if len(non_zero_indices) <= 2:
        return np.asarray([np.nan, np.nan, np.nan]), np.asarray(len(tac_times_in_minutes)*[np.nan])

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes[non_zero_indices],
                                        t_thresh_in_minutes=t_thresh_in_mins)

    if len(tac_times_in_minutes[non_zero_indices][t_thresh:]) <= 2:
        return np.asarray([np.nan, np.nan, np.nan]), np.asarray(len(tac_times_in_minutes)*[np.nan])

    y = cum_trapz(xdata=tac_times_in_minutes, ydata=tgt_tac_vals, initial=0.0)*weights
    y = y[non_zero_indices] / tgt_tac_vals[non_zero_indices]

    x1 = cum_trapz(xdata=tac_times_in_minutes, ydata=ref_tac_vals, initial=0.0)
    x1 = x1[non_zero_indices] / tgt_tac_vals[non_zero_indices]

    x2 = ref_tac_vals[non_zero_indices] / tgt_tac_vals[non_zero_indices]

    x_matrix = np.ones((len(y), 3), float)
    x_matrix[:, 0] = x1[:]*weights
    x_matrix[:, 1] = x2[:]*weights

    fit_ans = np.linalg.lstsq(x_matrix[t_thresh:], y[t_thresh:])[0]

    y_fit = x_matrix[:,0]*fit_ans[0] + x_matrix[:,1]*fit_ans[1] + x_matrix[:,2]*fit_ans[2]
    return fit_ans, y_fit


@numba.njit(fastmath=True)
def fit_mrtm_2003_to_tac(tac_times_in_minutes: np.ndarray,
                         tgt_tac_vals: np.ndarray,
                         ref_tac_vals: np.ndarray,
                         t_thresh_in_mins: float,
                         weights: np.ndarray=None):
    r"""
    Fit the 2003 Multilinear Reference Tissue Model (MRTM) to the provided target Time Activity
    Curve (TAC) values given the reference TAC, times, and threshold time (in minutes). The data
    are fit for all values beyond the threshold. We assume that the target TAC and reference TAC
    are sampled at the same times.

    .. important::
        This function assumes that both TACs are sampled at the same time, and that the time is in
        minutes.

    We have the following multilinear regression:

    .. math::

        C(T)=-\frac{V}{V^{\prime}b} \int_{0}^{T}C^{\prime}(t)\mathrm{d}t + \frac{1}{b}
        \int_{0}^{T}C(t)\mathrm{d}t - \frac{V}{V^{\prime}k_{2}^{\prime}b}C^{\prime}(T)


    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC values to fit the MRTM.
        ref_tac_vals (np.ndarray): Reference TAC values.
        t_thresh_in_mins (float): Threshold time in minutes.
        weights (np.ndarray): Weights applied to each frame. Default None.

    Returns:
        np.ndarray: Array containing fit results. (:math:`-\frac{V}{V^{\prime}b}`,
        :math:`\frac{1}{b}`, :math:`-\frac{V}{V^{\prime}k_{2}^{\prime}b}`)

    Note:
        This function is implemented with numba for improved performance.

    """
    if weights is None:
        weights = np.ones_like(tac_times_in_minutes)

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes,
                                        t_thresh_in_minutes=t_thresh_in_mins)
    if t_thresh == -1:
        return np.asarray([np.nan, np.nan, np.nan]), np.asarray(len(tac_times_in_minutes)*[np.nan])

    y = tgt_tac_vals*weights
    x_matrix = np.ones((len(y), 3), float)
    x_matrix[:, 0] = cum_trapz(xdata=tac_times_in_minutes, ydata=ref_tac_vals, initial=0.0)*weights
    x_matrix[:, 1] = cum_trapz(xdata=tac_times_in_minutes, ydata=tgt_tac_vals, initial=0.0)*weights
    x_matrix[:, 2] = ref_tac_vals*weights

    fit_ans = np.linalg.lstsq(x_matrix[t_thresh:], y[t_thresh:])[0]

    y_fit = x_matrix[:,0]*fit_ans[0] + x_matrix[:,1]*fit_ans[1] + x_matrix[:,2]*fit_ans[2]
    return fit_ans, y_fit


@numba.njit(fastmath=True)
def fit_mrtm2_2003_to_tac(tac_times_in_minutes: np.ndarray,
                          tgt_tac_vals: np.ndarray,
                          ref_tac_vals: np.ndarray,
                          t_thresh_in_mins: float,
                          k2_prime: float,
                          weights: np.ndarray=None):
    r"""
    Fit the second version of Multilinear Reference Tissue Model (MRTM2) to the provided target
    Time Activity Curve (TAC) values given the reference TAC, times, threshold time (in minutes),
    and k2_prime. The data are fit for all values beyond the threshold. We assume that the target
    TAC and reference TAC are sampled at the same times.

    .. important::
        This function assumes that both TACs are sampled at the same time, and that the time is in
        minutes.

    We have the following multilinear regression:

    .. math::

        C(T) = -\frac{V}{V^{\prime}b}\left(\int_{0}^{T}C^{\prime}(t)\mathrm{d}t
        -\frac{1}{k_{2}^{\prime}}C^{\prime}(T) \right)
        + \frac{1}{b} \int_{0}^{T}C(t)\mathrm{d}t


    Args:
        tac_times_in_minutes (np.ndarray): The array representing the time-points for both TACs.
        tgt_tac_vals (np.ndarray): Target TAC values to fit the MRTM2.
        ref_tac_vals (np.ndarray): Reference TAC values.
        t_thresh_in_mins (float): Threshold time in minutes.
        k2_prime (float): Kinetic parameter: washout rate for the reference region.
        weights (np.ndarray): Weights applied to each frame. Default None.

    Returns:
        np.ndarray: Array containing fit results. (:math:`-\frac{V}{V^{\prime}b}`,
            :math:`\frac{1}{b}`)

    Note:
        This function is implemented with numba for improved performance.

    """
    if weights is None:
        weights = np.ones_like(tac_times_in_minutes)

    t_thresh = get_index_from_threshold(times_in_minutes=tac_times_in_minutes,
                                        t_thresh_in_minutes=t_thresh_in_mins)
    if t_thresh == -1:
        return np.asarray([np.nan, np.nan]), np.asarray(len(tac_times_in_minutes)*[np.nan])

    x1 = cum_trapz(xdata=tac_times_in_minutes, ydata=ref_tac_vals, initial=0.0)
    x1 += ref_tac_vals / k2_prime
    x2 = cum_trapz(xdata=tac_times_in_minutes, ydata=tgt_tac_vals, initial=0.0)

    y = tgt_tac_vals*weights
    x_matrix = np.ones((len(y), 2), float)
    x_matrix[:, 0] = x1[:]*weights
    x_matrix[:, 1] = x2[:]*weights

    fit_ans = np.linalg.lstsq(x_matrix[t_thresh:], y[t_thresh:])[0]

    y_fit = x_matrix[:,0]*fit_ans[0] + x_matrix[:,1]*fit_ans[1]
    return fit_ans, y_fit


def calc_bp_from_mrtm_original_fit(fit_vals: np.ndarray) -> float:
    r"""
    Given the original MRTM (`Ichise et al., 1996`) fit values, we calculate the binding potential
    (BP).

    The binding potential (BP) is defined as:

    .. math::

        \mathrm{BP} = \beta_0 - 1

    where :math:`\beta_0` is the first fit coefficient.


    Args:
        fit_vals (np.ndarray): The multilinear regression fit values for the original MRTM.
            Output of :func:`fit_mrtm_original_to_tac`.

    Returns:
        float: Binding potential (BP) value.

    See Also:
        :func:`fit_mrtm_original_to_tac` where the order of the regression coefficients is laid out.

    """
    return fit_vals[0] - 1.0


def calc_bp_from_mrtm_2003_fit(fit_vals: np.ndarray) -> float:
    r"""
    Given the 2003 MRTM (`Ichise et al., 1996`) fit values, we calculate the binding potential
    (BP).

    The binding potential (BP) is calculated as:

    .. math::

        \mathrm{BP} = -\left(\frac{\beta_0}{\beta_1} + 1\right)

    where :math:`\beta_0` is the first fit coefficient, and :math:`\beta_1` is the second fit
    coefficient.

    Args:
        fit_vals (np.ndarray): The multilinear regression fit values for the 2003 MRTM.
            Output of :func:`fit_mrtm_2003_to_tac`.

    Returns:
        float: Binding potential (BP) value.

    See Also:
        :func:`fit_mrtm_2003_to_tac` where the order of the regression coefficients is laid out.

    """
    return -(fit_vals[0]/fit_vals[1] + 1.0)


def calc_bp_from_mrtm2_2003_fit(fit_vals: np.ndarray) -> float:
    r"""
    Given the 2003 MRTM2 (`Ichise et al., 1996`) fit values, we calculate the binding potential
    (BP).

    The binding potential (BP) is calculated as:

    .. math::

        \mathrm{BP} = -\left(\frac{\beta_0}{\beta_1} + 1\right)

    where :math:`\beta_0` is the first fit coefficient, and :math:`\beta_1` is the second fit
    coefficient.

    Args:
        fit_vals (np.ndarray): The multilinear regression fit values for the original MRTM.
            Output of :func:`fit_mrtm2_2003_to_tac`.

    Returns:
        float: Binding potential (BP) value.

    See Also:
        :func:`fit_mrtm2_2003_to_tac` where the order of the regression coefficients is laid out.

    """
    return -(fit_vals[0]/fit_vals[1] + 1.0)


def calc_k2prime_from_mrtm_original_fit(fit_vals: np.ndarray):
    r"""
    Given the original MRTM (`Ichise et al., 1996`) fit values, we calculate :math:`k_{2}^{\prime}`.

    The :math:`k_{2}^{\prime}` is calculated as:

    .. math::

         k_{2}^{\prime}= \frac{\beta_{0}}{\beta_{1}}

    where :math:`\beta_0` is the first fit coefficient and :math:`\beta_1` is the second fit
    coefficient.


    Args:
        fit_vals (np.ndarray): The multilinear regression fit values for the original MRTM.
            Output of :func:`fit_mrtm_original_to_tac`.

    Returns:
        float: :math:`k_2^\prime` value.

    See Also:
        :func:`fit_mrtm_original_to_tac` where the order of the regression coefficients is laid out.

    """
    return fit_vals[0]/fit_vals[1]


def calc_k2prime_from_mrtm_2003_fit(fit_vals: np.ndarray):
    r"""
    Given the 2003 MRTM (`Ichise et al., 2003`) fit values, we calculate :math:`k_{2}^{\prime}`.

    The :math:`k_{2}^{\prime}` is calculated as:

    .. math::

         k_{2}^{\prime}= \frac{\beta_{0}}{\beta_{2}}

    where :math:`\beta_0` is the first fit coefficient and :math:`\beta_2` is the third fit
    coefficient.


    Args:
        fit_vals (np.ndarray): The multilinear regression fit values for the original MRTM.
            Output of :func:`fit_mrtm_2003_to_tac`.

    Returns:
        float: :math:`k_2^\prime` value.

    See Also:
        :func:`fit_mrtm_2003_to_tac` where the order of the regression coefficients is laid out.

    """
    return fit_vals[0]/fit_vals[-1]



r"""
Command-line interface (CLI) for fitting Tissue Compartment Models (TCM) to PET Time-Activity Curves (TACs).

This module provides a CLI to interact with the tac_fitting module. It utilizes argparse to handle command-line
arguments.

The user must provide:
    * Input TAC file path
    * Region of Interest (ROI) TAC file path
    * Compartment model name for fitting. Supported models are '1tcm', '2tcm-k4zero', or 'serial-2tcm'.
    * Filename prefix for the output files
    * Output directory where the analysis results will be saved
    * Whether to ignore blood volume contributions while fitting
    * Threshold in minutes (input fitting threshold)

User can optionally provide:
    * Initial guesses for the fitting parameters
    * Lower and upper bounds for the fitting parameters
    * Maximum number of function iterations
    * Decay constant (in minutes) for per-frame weighting

This script utilizes the :class:`TCMAnalysis<petpal.tac_fitting.TCMAnalysis>` class to perform the TAC fitting and save
 the results accordingly.

Examples:
    In the proceeding example, we assume that we have an input TAC named 'input_tac.txt', and an ROI TAC named
    'roi_tac.tsv'. We want to try fitting a serial 2TCM to the ROI tac.
    
    .. code-block:: bash

        petpal-tcm-fit interp -i "input_tac.tsv"\
        -r "roi_tac.tsv"\
        -m "serial-2tcm"\
        -o "./" -p "cli_"\
        -t 35.0\
        -g 0.1 0.1 0.1 0.1 0.1\
        -l 0.0 0.0 0.0 0.0 0.0\
        -u 5.0 5.0 5.0 5.0 5.0\
        -f 2500 -n 4096 -b --print

    In this next example, we assume that we have an input TAC named 'input_tac.txt', an ROI TACs directory
     named './roi_tas/' which contains TACs for different ROIs, and a scan/image metadata file called 'scan_info.json'.
     We want to try fitting serial 2TCM to the ROI tacs, with the default bounds.

    .. code-block:: bash

        petpal-tcm-fit frame_avgd -i "input_tac.tsv"\
        -r "./roi_tacs/"\
        -s "scan_info.json"\
        -m "serial-2tcm"\
        -o "./" -p "cli_"\
        -f 2500 -n 4096 -b --print

See Also:
    :mod:`petpal.tac_fitting` - module for fitting TACs with TCMs.

"""

from typing import Union
import os
import numpy as np
import argparse
from ..kinetic_modeling import tac_fitting as pet_fit

_EXAMPLE_ = ('Fitting a TAC to the serial 2TCM using the interpolation strategy:\n\t'
             'petpal-tcm-fit interp -i "input_tac.txt"'
             ' -r "2tcm_tac.txt" '
             '-m "serial-2tcm" '
             '-o "./" -p "cli_" -t 35.0 '
             '-g 0.1 0.1 0.1 0.1 0.1 '
             '-l 0.0 0.0 0.0 0.0 0.0 '
             '-u 5.0 5.0 5.0 5.0 5.0 '
             '-f 2500 -n 4096 -b '
             '--print\n\n'
             'Fitting a TAC to the serial 2TCM using the frame averaged strategy:\n\t'
             'petpal-tcm-fit frame_avgd -i "input_tac.txt"'
             ' -r "2tcm_tac.txt" '
             '-s scan_metadata.json '
             '-m "serial-2tcm" '
             '-o "./" -p "cli_" '
             '-g 0.1 0.1 0.1 0.1 0.1 '
             '-l 0.0 0.0 0.0 0.0 0.0 '
             '-u 5.0 5.0 5.0 5.0 5.0 '
             '-f 2500 -n 4096 -b '
             '--print'
             )

def add_common_io_args(parser: argparse.ArgumentParser) -> argparse._ArgumentGroup:
    r"""Adds common input/output arguments to the provided argument parser.

    This helper function configures arguments that are shared between all subcommands
    for specifying input TAC paths, ROI TAC paths, output directory and output filename prefix.

    Args:
        parser (argparse.ArgumentParser): Argument parser (or subparser) to which the
            common I/O arguments will be added.

    Returns:
        argparse._ArgumentGroup: The argument group containing the I/O arguments.
    """
    grp_io = parser.add_argument_group('IO Paths and Prefixes')
    grp_io.add_argument("-i", "--input-tac-path", required=True, help="Path to the input TAC file.")
    grp_io.add_argument("-r", "--roi-tac-path", required=True, help="Path to the ROI TAC file. "
                                                                    "Can also be the path to a directory where each "
                                                                    "TAC in the directory corresponds to an ROI to be "
                                                                    "analyzed.")
    grp_io.add_argument("-o", "--output-directory", required=True, help="Path to the output directory.")
    grp_io.add_argument("-p", "--output-filename-prefix", required=True,
                        help="Prefix for the output filenames. Typically 'sub-xxxx_ses-xx'")
    return grp_io


def add_common_analysis_args(parser: argparse.ArgumentParser):
    r"""Adds common analysis-related arguments to the provided argument parser.

    This helper function configures arguments controlling model selection, parameter
    initialization, parameter bounds, maximum number of fitting iterations and
    resampling behavior.

    Args:
        parser (argparse.ArgumentParser): Argument parser (or subparser) to which the
            analysis arguments will be added.
    """
    grp_analysis = parser.add_argument_group('Analysis Parameters')
    grp_analysis.add_argument("-m", "--model", required=True,
                              choices=['1tcm', '2tcm-k4zero', 'serial-2tcm', '2tcm'],
                              help="Analysis model to fit.")
    grp_analysis.add_argument("-g", "--initial-guesses", required=False, nargs='+', type=float,
                              help="Initial guesses for each fitting parameter.")
    grp_analysis.add_argument("-l", "--lower-bounds", required=False, nargs='+', type=float,
                              help="Lower bounds for each fitting parameter.")
    # grp_analysis.add_argument("-w", "--weighting-decay-constant", required=False, type=float, default=None,
    #                           help="Decay constant for computing per-frame weighting for fits.")
    grp_analysis.add_argument("-u", "--upper-bounds", required=False, nargs='+', type=float,
                              help="Upper bounds for each fitting parameter.")
    grp_analysis.add_argument("-f", "--max-fit-iterations", required=False, default=2500, type=int,
                              help="Maximum number of function iterations")
    grp_analysis.add_argument("-n", "--resample-num", required=False, default=4096, type=int,
                              help="Number of samples for uniform linear interpolation of provided TACs.")


def add_common_print_args(parser: argparse.ArgumentParser):
    r"""Adds common verbosity/printing arguments to the provided argument parser.

    This helper function configures arguments related to optional printing of analysis
    results to the console.

    Args:
        parser (argparse.ArgumentParser): Argument parser (or subparser) to which the
            printing/verbosity arguments will be added.
    """
    grp_verbose = parser.add_argument_group('Additional Options')
    grp_verbose.add_argument("--print", action="store_true", help="Whether to print the analysis results.")


def _generate_args() -> argparse.Namespace:
    r"""Generates and handles the arguments for the command-line interface.

    This function sets up the argument parser, adds required and optional arguments, and parses input arguments.

    Returns:
        argparse.Namespace: Parsed command-line arguments.

    Raises:
        argparse.ArgumentError: If necessary arguments are missing or invalid arguments are provided.
    """
    parser = argparse.ArgumentParser(prog='petpal-tcm-fit',
                                     description='Command line interface for fitting Tissue Compartment Models (TCM) '
                                                 'to PET Time Activity Curves (TACs).',
                                     formatter_class=argparse.RawTextHelpFormatter, epilog=_EXAMPLE_)

    subparsers = parser.add_subparsers(dest='strategy', description='Strategy for fitting TACs.',
                                       help='Strategy for fitting TACs.')

    frame_avgd_parser = subparsers.add_parser('frame_avgd', help='Perform analysis on properly frame averaged '
                                                                 'TACs using scan information. Usually more stable to '
                                                                 'noise, and the preferred strategy for non-simulated TACs')
    interpolated_parser = subparsers.add_parser('interp', help='Perform analysis on interpolated TACs '
                                                               'without using scan information.')

    sub_parser_list = [frame_avgd_parser, interpolated_parser]

    for a_parser in sub_parser_list:
        match a_parser.prog.split(' ')[-1]:
            case 'interp':
                grp_io = add_common_io_args(a_parser)
                grp_io.add_argument("-t", "--input-fitting-threshold-in-mins", required=False, type=float, default=30.0,
                                    help="Threshold in minutes for fitting the later half of the input function.")
                grp_io.add_argument("-b", "--ignore-blood-volume", required=False,
                                    default=False, action='store_true',
                                    help="Whether to ignore any blood volume contributions while fitting.")
            case 'frame_avgd':
                grp_io = add_common_io_args(a_parser)
                grp_io.add_argument('-s', '--scan-metadata-path', required=True, type=str,
                                    help="Path to the scan metadata file. Can also be the path to the .nii.gz file"
                                         "if the metadata shares the scan name: `*.nii.gz` -> `*json`.")
            case _:
                pass
        add_common_analysis_args(a_parser)
        add_common_print_args(a_parser)

    return parser.parse_args()


def _generate_bounds(initial: Union[list, None],
                     lower: Union[list, None],
                     upper: Union[list, None]) -> Union[np.ndarray, None]:
    r"""Generates the bounds for the fitting parameters.

    This function takes lists of initial fitting parameters, lower bounds, and upper bounds.
    All lists must have the same length. If no initial parameters are provided, the function returns None.

    Args:
        initial (list, optional): List of initial guesses for fitting parameters. If None, no bounds are generated.
        lower (list, optional): List of lower bounds for fitting parameters.
        upper (list, optional): List of upper bounds for fitting parameters.

    Returns:
        (np.ndarray, optional): If initial is not None, then we return a numpy array of shape [n, 3],
        where n is the number of parameters, where column 0 has the initial guesses, column 1 has lower bounds, and
        column 2 has upper bounds. If initial is None, function will return None.

    Raises:
        ValueError: If initial is not None and the length of initial, lower, and upper are not the same.
    """
    if initial is not None:
        if (len(initial) != len(lower)) or (len(initial) != len(upper)) or (len(upper) != len(lower)):
            raise ValueError("The number of initial guesses, lower bounds and upper bounds must be the same.")
        return np.asarray(np.asarray([initial, lower, upper]).T)
    else:
        return None


def fit_props_printer(fit_props: dict, segment: str | None = None) -> None:
    r"""Nicely formats and prints fit parameter estimates and uncertainties.

    This function prints a table of fitted parameter values, standard errors and
    percent errors. If a segment label is provided, it is printed as a header
    before the table, which is useful when printing results for multiple ROIs.

    Args:
        fit_props (dict): Dictionary containing the fit results. It is expected to
            contain a ``"FitProperties"`` entry with nested ``"FitValues"`` and
            ``"FitStdErr"`` mappings of parameter names to numerical values.
        segment (str, optional): Optional label for the TAC segment (for example,
            ROI name). If provided, the label is printed above the table.
    """
    title_str = f"{'Param':<5} {'FitVal':<6}    {'StdErr':<8} ({'%Err':>6})|"
    title_len = len(title_str)
    if segment :
        print(f'{"Segment: " + segment:#^{title_len}}')
    print("-" * title_len)
    print(title_str)
    print("-" * title_len)
    vals = fit_props["FitProperties"]["FitValues"]
    errs = fit_props["FitProperties"]["FitStdErr"]
    for param_name in vals:
        val = vals[param_name]
        err = errs[param_name]
        percent_err = f"{err / val * 100:>5.2f}%" if val > 0 else "NAN"
        print(f"{param_name:<5} {val:<6.4f} +- {err:<8.4f} ({percent_err:>6})|")

    print("-" * title_len)

def main():
    args = _generate_args()
    
    bounds = _generate_bounds(initial=args.initial_guesses, lower=args.lower_bounds, upper=args.upper_bounds)

    if args.strategy is None:
        args.print_help()
        raise ValueError("No fitting strategy specified!")

    common_kwargs = dict(input_tac_path=args.input_tac_path,
                         roi_tac_path=args.roi_tac_path,
                         roi_tacs_dir=args.roi_tac_path,
                         output_directory=args.output_directory,
                         output_filename_prefix=args.output_filename_prefix,
                         compartment_model=args.model,
                         parameter_bounds=bounds,
                         weights=None,
                         resample_num=args.resample_num, )

    is_single_tac = os.path.isfile(args.roi_tac_path)
    common_kwargs.pop('roi_tacs_dir') if is_single_tac else common_kwargs.pop('roi_tac_path')


    if args.strategy == 'frame_avgd':
        strategy_kwargs = common_kwargs | dict(scan_info_path=args.scan_metadata_path)
        AnalysisClass = pet_fit.FrameAveragedTCMAnalysis if is_single_tac else pet_fit.FrameAveragedMultiTACTCMAnalysis
    else:
        strategy_kwargs = common_kwargs | dict(aif_fit_thresh_in_mins=args.input_fitting_threshold_in_mins,
                                               max_func_iters=args.max_fit_iterations,
                                               ignore_blood_volume=args.ignore_blood_volume)
        AnalysisClass = pet_fit.TCMAnalysis if is_single_tac else pet_fit.MultiTACTCMAnalysis

    os.makedirs(args.output_directory, exist_ok=True)
    tac_fitting = AnalysisClass(**strategy_kwargs)
    tac_fitting.run_analysis()
    tac_fitting.save_analysis()
    
    if args.print:
        if os.path.isfile(args.roi_tac_path):
            fit_props_printer(fit_props=tac_fitting.analysis_props)
        else:
            for seg_name, seg_props in zip(tac_fitting.inferred_seg_labels, tac_fitting.analysis_props):
                fit_props_printer(fit_props=seg_props, segment=seg_name)


if __name__ == "__main__":
    main()



import os
import warnings
import pytest
import pandas as pd

from petpal.utils.image_io import flatten_metadata
import petpal.kinetic_modeling.graphical_analysis as ga

def _make_instance(rsquared):
    inst = ga.MultiTACGraphicalAnalysis.__new__(ga.MultiTACGraphicalAnalysis)
    inst.analysis_props = [{'RSquared': rsquared}]
    inst.output_directory = "/tmp"
    inst.output_filename_prefix = "prefix"
    inst.method = "patlak"
    inst.inferred_seg_labels = ["roi1", "roi2"]
    return inst


def test_save_analysis_raises_if_run_not_called():
    inst = _make_instance(rsquared=None)
    with pytest.raises(RuntimeError):
        inst.save_analysis()

def test_save_analysis_calls_tsv_and_json(monkeypatch):
    inst = _make_instance(rsquared=0.95)
    calls = {"tsv": 0, "json": 0}

    def km_multifit_analysis_to_tsv_without_save_file(analysis_props: list[dict],
                                    output_directory: str,
                                    output_filename_prefix: str,
                                    method: str,
                                    inferred_seg_labels: list[str]):
        filename = f'{output_filename_prefix}_desc-{method}_fitprops.tsv'
        filepath = os.path.join(output_directory, filename)
        fit_table = pd.DataFrame()
        for seg_name, fit_props in zip(inferred_seg_labels, analysis_props):
            tmp_table = pd.DataFrame(flatten_metadata(fit_props),index=[seg_name])
            fit_table = pd.concat([fit_table,tmp_table])
        calls["tsv"] += 1
        assert isinstance(analysis_props, list)
        assert isinstance(output_directory, str)
        assert isinstance(output_filename_prefix, str)
        assert isinstance(method, str)
        assert isinstance(inferred_seg_labels, list)

    def km_multifit_analysis_to_jsons_without_save_file(analysis_props: list[dict],
                                    output_directory: str,
                                    output_filename_prefix: str,
                                    method: str,
                                    inferred_seg_labels: list[str]):
        for seg_name, fit_props in zip(inferred_seg_labels, analysis_props):
            filename = [output_filename_prefix,
                        f'desc-{method}',
                        f'seg-{seg_name}',
                        'fitprops.json']
            filename = '_'.join(filename)
            filepath = os.path.join(output_directory, filename)
        calls["json"] += 1
        assert isinstance(analysis_props, list)
        assert isinstance(output_directory, str)
        assert isinstance(output_filename_prefix, str)
        assert isinstance(method, str)
        assert isinstance(inferred_seg_labels, list)

    monkeypatch.setattr(ga, "km_multifit_analysis_to_tsv", km_multifit_analysis_to_tsv_without_save_file)
    monkeypatch.setattr(ga, "km_multifit_analysis_to_jsons", km_multifit_analysis_to_jsons_without_save_file)

    # Default behavior: TSV only
    calls["tsv"] = calls["json"] = 0
    inst.save_analysis(output_as_tsv=True, output_as_json=False)
    assert calls["tsv"] == 1 and calls["json"] == 0

    # JSON only
    calls["tsv"] = calls["json"] = 0
    inst.save_analysis(output_as_tsv=False, output_as_json=True)
    assert calls["tsv"] == 0 and calls["json"] == 1

    # Both
    calls["tsv"] = calls["json"] = 0
    inst.save_analysis(output_as_tsv=True, output_as_json=True)
    assert calls["tsv"] == 1 and calls["json"] == 1

def test_save_analysis_warns_when_no_output_requested(monkeypatch):
    inst = _make_instance(rsquared=0.5)
    # prevent actual functions being called if mistakenly invoked
    monkeypatch.setattr(ga, "km_multifit_analysis_to_tsv", lambda *a, **k: (_ for _ in ()).throw(AssertionError("should not be called")))
    monkeypatch.setattr(ga, "km_multifit_analysis_to_jsons", lambda *a, **k: (_ for _ in ()).throw(AssertionError("should not be called")))

    with pytest.warns(UserWarning):
        inst.save_analysis(output_as_tsv=False, output_as_json=False)


from . import steps_base
from . import steps_containers
from . import preproc_steps
from . import kinetic_modeling_steps
from . import pca_guided_idif_steps
from . import pipelines

def main():
    print("PETPAL - Pipelines")


if __name__ == "__main__":
    main()



"""
This module contains a collection of functions to compute Time-Activity Curves (TACs) for common Tissue Compartment
Models (TCMs). These models are commonly used for kinetic analysis of PET TACs.

Note:
    All response functions in this module are decorated with :func:`numba.jit`. It compiles the function to
    machine code at runtime (Just-In-Time compilation), which usually provides a significant speed-up.

Requires:
    The module relies on the :doc:`numpy <numpy:index>` and :doc:`numba <numba:index>` modules.

TODO:
    Add the derivations of the solutions to the Tissue Compartment Models in the module docstring.
    
"""

import numba
import numpy as np
from scipy.signal import convolve as sp_conv
import lmfit

from ..utils.time_activity_curve import get_frame_averaged_tac_vals


def calc_convolution_with_check(f: np.ndarray, g: np.ndarray, dt: float) -> np.ndarray:
    r"""Performs a discrete convolution of two arrays, assumed to represent time-series data. Checks if the arrays are
    of the same shape.
    
    Let ``f``:math:`=f(t)` and ``g``:math:`=g(t)` where both functions are 0 for :math:`t\leq0`. Then,
    the output, :math:`h(t)`, is
    
    .. math::
    
        h(t) = \int_{0}^{t}f(s)g(s-t)\mathrm{d}s
    
    Args:
        f (np.ndarray): Array containing the values for the input function.
        g (np.ndarray): Array containing values for the response function.
        dt (np.ndarray): The step-size, in the time-domain, between samples for ``f`` and ``g``.

    Returns:
        (np.ndarray): Convolution of the two arrays scaled by ``dt``.
        
    .. important::
        This function does not use :func:`numba.jit`. Therefore, it cannot be used directly inside JIT'ed functions.
        
    """
    assert len(f) == len(g), f"The provided arrays must have the same lengths! f:{len(f):<6} and g:{len(g):<6}."
    vals = sp_conv(f, g, mode='full')
    return vals[:len(f)] * dt


@numba.njit(fastmath=True, cache=True)
def discrete_convolution_with_exponential(func_times: np.ndarray, func_vals: np.ndarray, k1: float, k2: float):
    r"""Computes the convolution of the given function with an exponential kernel.

    Given the provided function :math:`u(t)`, sampled evenly with respect to time, we calculate
    :math:`c(t) = k_{1} \exp(-k_{2} t) \otimes u(t)` where :math:`\otimes` represents
    the convolution operator. This implementation is :math:`\mathcal{O}(N)` due to the
    simple recurrence relationship arising from an exponential kernel. This implementation is based
    heavily on the `KMAPLIB <https://github.com/ShareKM/KMAP-C/blob/caeb10b7bec1e841132f879856949f00a9624883/src/kmaplib_common.cpp#L43-L70>`_
    implementation

    .. important::
        The function assumes that the times are evenly sampled. Answers will be incorrect if this is not the case.

    Args:
        func_vals (np.ndarray): Array containing function values for :math:`t\geq0`.
            Assumed to be evenly sampled with respect to :math:`t`.
        func_times (np.ndarray): Array containing time-points where :math:`t\geq0`.
            Assumed to be evenly sampled with respect to :math:`t`.
        k1 (float): Rate constant for transport from first tissue compartment.
        k2 (float): Rate constant for transport from second tissue compartment.

    Returns:
        (np.ndarray): Array containing the convolution of an exponential function with the provided function.

    See Also:
        :func:`calc_convolution_with_check` for a more general convolution function
    """
    dt = func_times[1] - func_times[0]
    num_times = len(func_times)
    c_out = np.zeros(num_times)

    prev = 0
    if k2 <= 1e-8:
        for i in range(0, num_times):
            prev += func_vals[i]
            c_out[i] = prev
        return k1 * c_out * dt
    else:
        _k1 = k1 * dt
        _k2 = k2 * dt
        ek2 = np.exp(-_k2)
        tmp = _k1 * (1.0 - ek2) / _k2
        u_tmp = tmp * func_vals
        for i in range(0, num_times):
            prev = prev * ek2 + u_tmp[i]
            c_out[i] = prev
        return c_out

@numba.njit()
def response_function_1tcm_c1(t: np.ndarray, k1: float, k2: float) -> np.ndarray:
    r"""The response function for the 1TCM :math:`f(t)=k_1 e^{-k_{2}t}`
    
    Args:
        t (np.ndarray): Array containing time-points where :math:`t\geq0`.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from tissue compartment back to plasma/blood.

    Returns:
        (np.ndarray): Array containing response function values given the constants.
    """
    return k1 * np.exp(-k2 * t)


@numba.njit()
def response_function_2tcm_with_k4zero_c1(t: np.ndarray, k1: float, k2: float, k3: float) -> np.ndarray:
    r"""
    The response function for first compartment in the serial 2TCM with
    :math:`k_{4}=0`; :math:`f(t)=k_{1}e^{-(k_{2} + k_{3})t}`.
    
    Args:
        t (np.ndarray): Array containing time-points where :math:`t\geq0`.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from tissue compartment to irreversible compartment.

    Returns:
        (np.ndarray): Array containing response function values for first compartment given the constants.
        
    See Also:
        :func:`response_function_2tcm_with_k4zero_c2`
        
    """
    return k1 * np.exp(-(k2 + k3) * t)


@numba.njit()
def response_function_2tcm_with_k4zero_c2(t: np.ndarray, k1: float, k2: float, k3: float) -> np.ndarray:
    r"""The response function for second compartment in the serial 2TCM with :math:`k_{4}=0`;
    :math:`f(t)=\frac{k_{1}k_{3}}{k_{2}+k_{3}}(1-e^{-(k_{2} + k_{3})t})`.

    Args:
        t (np.ndarray): Array containing time-points where :math:`t\geq0`.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from tissue compartment to irreversible compartment.

    Returns:
        (np.ndarray): Array containing response function values for first compartment given the constants.
    
    See Also:
        :func:`response_function_2tcm_with_k4zero_c1`
    """
    return ((k1 * k3) / (k2 + k3)) * (1.0 - np.exp(-(k2 + k3) * t))


@numba.njit()
def response_function_serial_2tcm_c1(t: np.ndarray, k1: float, k2: float, k3: float, k4: float) -> np.ndarray:
    r"""The response function for first compartment in the *serial* 2TCM.
    
    .. math::
    
        f(t) = \frac{k_{1}}{\Delta \alpha} \left[ (k_{4}-\alpha_{1})e^{-\alpha_{1}t} + (\alpha_{2}-k_{4})e^{-\alpha_{2}t}\right]
    
    where
    
    .. math::
    
        a&= k_{2}+k_{3}+k_{4}\\
        \alpha_{1}&=\frac{a-\sqrt{a^{2}-4k_{2}k_{4}}}{2}\\
        \alpha_{1}&=\frac{a+\sqrt{a^{2}-4k_{2}k_{4}}}{2}\\
        \Delta \alpha&=\alpha_2 - \alpha_1
    
    Args:
        t (np.ndarray): Array containing time-points where :math:`t\geq0`.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from first tissue compartment to second tissue compartment.
        k4 (float): Rate constant for transport from second tissue compartment back to first tissue compartment.

    Returns:
        (np.ndarray): Array containing response function values for first compartment given the constants.
        
    See Also:
        * :func:`response_function_serial_2tcm_c2`
        * :func:`response_function_2tcm_with_k4zero_c1` for when :math:`k_{4}=0` (irreversible second compartment).
        
    """
    a = k2 + k3 + k4
    alpha_1 = (a - np.sqrt((a ** 2.) - 4.0 * k2 * k4)) / 2.0
    alpha_2 = (a + np.sqrt((a ** 2.) - 4.0 * k2 * k4)) / 2.0
    delta_a = alpha_2 - alpha_1
    
    return (k1 / delta_a) * ((k4 - alpha_1) * np.exp(-alpha_1 * t) + (alpha_2 - k4) * np.exp(-alpha_2 * t))


@numba.njit()
def response_function_serial_2tcm_c2(t: np.ndarray, k1: float, k2: float, k3: float, k4: float) -> np.ndarray:
    r"""The response function for second compartment in the *serial* 2TCM.

    .. math::
    
        f(t) = \frac{k_{1}k_{3}}{a} \left[ e^{-\alpha_{1}t} - e^{-\alpha_{2}t}\right]

    where

    .. math::
    
        a&= k_{2}+k_{3}+k_{4}\\
        \alpha_{1}&=\frac{a-\sqrt{a^{2}-4k_{2}k_{4}}}{2}\\
        \alpha_{1}&=\frac{a+\sqrt{a^{2}-4k_{2}k_{4}}}{2}\\
        \Delta \alpha&=\alpha_2 - \alpha_1

    Args:
        t (np.ndarray): Array containing time-points where :math:`t\geq0`.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from first tissue compartment to second tissue compartment.
        k4 (float): Rate constant for transport from second tissue compartment back to first tissue compartment.

    Returns:
        (np.ndarray): Array containing response function values for second compartment given the constants.
        
    See Also:
        * :func:`response_function_serial_2tcm_c2`
        * :func:`response_function_2tcm_with_k4zero_c2` for when :math:`k_{4}=0` (irreversible second compartment).
        
    """
    a = k2 + k3 + k4
    alpha_1 = (a - np.sqrt((a ** 2.) - 4.0 * k2 * k4)) / 2.0
    alpha_2 = (a + np.sqrt((a ** 2.) - 4.0 * k2 * k4)) / 2.0
    delta_a = alpha_2 - alpha_1
    
    return (k1 * k3 / delta_a) * (np.exp(-alpha_1 * t) - np.exp(-alpha_2 * t))

@numba.njit(cache=True, fastmath=True)
def gen_tac_1tcm_cpet_from_tac(tac_times: np.ndarray,
                               tac_vals: np.ndarray,
                               k1: float,
                               k2: float,
                               vb: float = 0.0) -> list[np.ndarray, np.ndarray]:
    r"""Calculate the TTAC, given the input TAC, for a 1TCM as an explicit convolution.

    .. important::
        This function assumes that the provided input TAC is sampled evenly with respect to time.

    Computes the following:


    .. math::

        C_\mathrm{T} = C_{P}(t) \otimes k_{1}e^{-k_{2}t}


    where :math:`C_\mathrm{T}` is the output TTAC, and :math:`C_{P}(t)` is the input/plasma TAC


    Args:
        tac_times (np.ndarray): Array containing time-points where :math:`t\geq0` and equal time-steps.
        tac_vals (np.ndarray): Array containing TAC activities.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.

    Returns:
        ((np.ndarray, np.ndarray)): Arrays containing the times and TTAC given the input TAC and parameters.
    """
    
    c1 = discrete_convolution_with_exponential(func_times=tac_times, func_vals=tac_vals, k1=k1, k2=k2)
    return [tac_times, (1.0-vb)*c1 + vb*tac_vals]


def generate_tac_2tcm_with_k4zero_c1_from_tac(tac_times: np.ndarray,
                                              tac_vals: np.ndarray,
                                              k1: float,
                                              k2: float,
                                              k3: float) -> np.ndarray:
    r"""
    Calculate the TTAC of the first comparment, given the input TAC, for a 2TCM (with :math:`k_{4}=0`) as an explicit
    convolution.
    
    Args:
        tac_times (np.ndarray): Array containing time-points where :math:`t\geq0` and equal time-steps.
        tac_vals (np.ndarray): Array containing TAC activities.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from tissue compartment to irreversible compartment.

    Returns:
        ((np.ndarray, np.ndarray)): Arrays containing the times and TTAC given the input TAC and parameters.
        
    See Also:
        :func:`response_function_2tcm_with_k4zero_c1` for more details about the 2TCM response function, of the first
        compartment, used for the convolution.
        
    """
    _resp_vals = response_function_2tcm_with_k4zero_c1(t=tac_times, k1=k1, k2=k2, k3=k3)
    dt = tac_times[1] - tac_times[0]
    c1 = calc_convolution_with_check(f=tac_vals, g=_resp_vals, dt=dt)
    return np.asarray([tac_times, c1])


def generate_tac_2tcm_with_k4zero_c2_from_tac(tac_times: np.ndarray,
                                              tac_vals: np.ndarray,
                                              k1: float,
                                              k2: float,
                                              k3: float) -> np.ndarray:
    r"""
    Calculate the TTAC of the second comparment, given the input TAC, for a 2TCM (with :math:`k_{4}=0`) as an
    explicit convolution.
    
    Args:
        tac_times (np.ndarray): Array containing time-points where :math:`t\geq0` and equal time-steps.
        tac_vals (np.ndarray): Array containing TAC activities.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from tissue compartment to irreversible compartment.

    Returns:
        ((np.ndarray, np.ndarray)): Arrays containing the times and TTAC given the input TAC and parameters.
        
    See Also:
        :func:`response_function_2tcm_with_k4zero_c2` for more details about the 2TCM response function, of the second
        compartment, used for the convolution.
        
    """
    _resp_vals = response_function_2tcm_with_k4zero_c2(t=tac_times, k1=k1, k2=k2, k3=k3)
    dt = tac_times[1] - tac_times[0]
    c2 = calc_convolution_with_check(f=tac_vals, g=_resp_vals, dt=dt)
    return np.asarray([tac_times, c2])


def gen_tac_2tcm_with_k4zero_cpet_from_tac(tac_times: np.ndarray,
                                           tac_vals: np.ndarray,
                                           k1: float,
                                           k2: float,
                                           k3: float,
                                           vb: float = 0.0) -> list[np.ndarray, np.ndarray]:
    r"""
    Calculate the PET-TTAC (sum of both compartments), given the input TAC, for a 2TCM (with :math:`k_{4}=0`) as an
    explicit convolution.
    
    Args:
        tac_times (np.ndarray): Array containing time-points where :math:`t\geq0` and equal time-steps.
        tac_vals (np.ndarray): Array containing TAC activities.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from tissue compartment to irreversible compartment.

    Returns:
        ((np.ndarray, np.ndarray)): Arrays containing the times and TTAC given the input TAC and parameters.
        
    See Also:
        * :func:`gen_tac_2tcm_cpet_from_tac` for more details about the 2TCM function.

    """
    return gen_tac_2tcm_cpet_from_tac(tac_times=tac_times, tac_vals=tac_vals, k1=k1, k2=k2, k3=k3, k4=0, vb=vb)


def generate_tac_serial_2tcm_c1_from_tac(tac_times: np.ndarray,
                                         tac_vals: np.ndarray,
                                         k1: float,
                                         k2: float,
                                         k3: float,
                                         k4: float) -> np.ndarray:
    r"""
    Calculate the TTAC of the first compartment, given the input TAC, for a serial 2TCM as an explicit convolution.
    
    Args:
        tac_times (np.ndarray): Array containing time-points where :math:`t\geq0` and equal time-steps.
        tac_vals (np.ndarray): Array containing TAC activities.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from tissue compartment to second compartment.
        k4 (float): Rate constant for transport from second tissue compartment back to first tissue compartment.

    Returns:
        ((np.ndarray, np.ndarray)): Arrays containing the times and TTAC given the input TAC and parameters.
        
    See Also:
        * :func:`response_function_serial_2tcm_c1` for more details about the 2TCM response function, of the first
            compartment, used for the convolution.
        * :func:`response_function_2tcm_with_k4zero_c1` for more details about the 2TCM response function
            (with :math:`k_{4}=0`), of the first compartment, used for the convolution.
        
    """
    _resp_vals = response_function_serial_2tcm_c1(t=tac_times, k1=k1, k2=k2, k3=k3, k4=k4)
    dt = tac_times[1] - tac_times[0]
    c1 = calc_convolution_with_check(f=tac_vals, g=_resp_vals, dt=dt)
    return np.asarray([tac_times, c1])


def generate_tac_serial_2tcm_c2_from_tac(tac_times: np.ndarray,
                                         tac_vals: np.ndarray,
                                         k1: float,
                                         k2: float,
                                         k3: float,
                                         k4: float) -> np.ndarray:
    r"""
    Calculate the TTAC of the second comparment, given the input TAC, for a serial 2TCM as an explicit convolution.

    Args:
        tac_times (np.ndarray): Array containing time-points where :math:`t\geq0` and equal time-steps.
        tac_vals (np.ndarray): Array containing TAC activities.
        k1 (float): Rate constant for transport from plasma/blood to tissue compartment.
        k2 (float): Rate constant for transport from first tissue compartment back to plasma/blood.
        k3 (float): Rate constant for transport from tissue compartment to second compartment.
        k4 (float): Rate constant for transport from second tissue compartment back to first tissue compartment.

    Returns:
        ((np.ndarray, np.ndarray)): Arrays containing the times and TTAC given the input TAC and parameters.

    See Also:
        * :func:`response_function_serial_2tcm_c2` for more details about the 2TCM response function, of the second
            compartment, used for the convolution.
        * :func:`response_function_2tcm_with_k4zero_c2` for more details about the 2TCM response function
            (with :math:`k_{4}=0`), of the second compartment, used for the convolution.
        
    """
    _resp_vals = response_function_serial_2tcm_c2(t=tac_times, k1=k1, k2=k2, k3=k3, k4=k4)
    dt = tac_times[1] - tac_times[0]
    c2 = calc_convolution_with_check(f=tac_vals, g=_resp_vals, dt=dt)
    return np.asarray([tac_times, c2])


@numba.njit(fastmath=True, cache=True)
def gen_tac_2tcm_cpet_from_tac(tac_times: np.ndarray,
                               tac_vals: np.ndarray,
                               k1: float,
                               k2: float,
                               k3: float,
                               k4: float,
                               vb: float = 0.0) -> list[np.ndarray, np.ndarray]:
    r"""Generates PET TAC values using the serial 2TCM.

    Computes the PET TAC, given the input TAC, using a serial Two Tissue Compartment Model (2TCM).
    Since this function uses :func:`discrete_convolution_with_exponential`, this function
    is also JIT'ed.

    .. important:
        This function assumes that the provided input TAC is sampled evenly with respect to time.


    We calculate

    .. math::

        C_\mathrm{T} = (1-v_{\mathrm{B}}) (C_{1} + C_{2}) + v_{\mathrm{B}} C_\mathrm{P}


    where :math:`C_\mathrm{T}` is the output PET TAC, :math:`v_{\mathrm{B}}` is the blood volume,
    :math:`C_{1}` and :math:`C_{2}` are the TACs for the first and second tissue compartments, and
    :math:`C_\mathrm{P}` is the input/plasma TAC. Furthermore,

    .. math::

        \begin{align}
        C_{1} &= \frac{k_{1}(k_{4} - \alpha_{2})}{\beta}C_{a} + \frac{k_{1}(\alpha_{1} - k_{4})}{\beta}C_{b}\\
        C_{2} &= \frac{k_{1}k_{3}}{\beta}C_{a} - \frac{k_{1}k_{3}}{\beta}C_{b}
        \end{align}

    where

    .. math::

        \begin{align}
        C_{a}&=C_\mathrm{P}(t) \otimes e^{-\alpha_{2}\times t}\\
        C_{b}&=C_\mathrm{P}(t) \otimes e^{-\alpha_{1}\times t}\\
        \beta&= \sqrt{x^{2}-4k_{2}k_{4}}\\
        \alpha_{1}&=\frac{x-\beta}{2}\\
        \alpha_{2}&=\frac{x+\beta}{2}\\
        x&= k_{2}+k_{3}+k_{4}
        \end{align}


    Args:
        tac_times (np.ndarray): Time points for the input TAC values.
        tac_vals (np.ndarray): TAC values corresponding to the time points, assumed to be in minutes.
        k1 (float): Rate constant for blood-to-first-tissue transport.
        k2 (float): Rate constant for first-tissue-to-blood transport.
        k3 (float): Rate constant for 1st compartment to 2nd compartment transport.
        k4 (float): Rate constant for 2nt compartment to 1st compartment transport.
        vb (float, optional): Vascular blood fraction. Defaults to 0.0.

    Returns:
        np.ndarray: Simulated PET TAC computed using the serial 2TCM.
    """
    k234 = k2 + k3 + k4
    beta = np.sqrt(k234 * k234 - 4.0 * k2 * k4)
    a1 = (k234 - beta) / 2.0
    a2 = (k234 + beta) / 2.0

    c_a1 = discrete_convolution_with_exponential(func_times=tac_times, func_vals=tac_vals, k1=1.0, k2=a1)
    c_a2 = discrete_convolution_with_exponential(func_times=tac_times, func_vals=tac_vals, k1=1.0, k2=a2)

    if beta <= 1.0e-8:
        f1 = 0.0
        f2 = 0.0
        b1 = 0.0
        b2 = 0.0
    else:
        f1 = k1 / beta * (k4 - a1)
        f2 = k1 / beta * (a2 - k4)
        b1 = k1 / beta * k3
        b2 = -b1

    c_1 = f1 * c_a1 + f2 * c_a2
    c_2 = b1 * c_a1 + b2 * c_a2
    c_pet = (1.0 - vb) * (c_1 + c_2) + vb * tac_vals

    return [tac_times, c_pet]


def model_serial_1tcm_frame_avgd(params: lmfit.Parameters,
                                 cp_times: np.ndarray,
                                 cp_vals: np.ndarray,
                                 frame_idx_pairs: np.ndarray,
                                 data: np.ndarray | None = None,
                                 eps: np.ndarray | float | None = None) -> np.ndarray:
    par_vals = params.valuesdict()
    model = gen_tac_1tcm_cpet_from_tac(tac_times=cp_times,
                                       tac_vals=cp_vals,
                                       k1=par_vals['k1'],
                                       k2=par_vals['k2'],
                                       vb=par_vals['vb'])[-1]

    model = get_frame_averaged_tac_vals(tac_vals=model, frame_idx_pairs=frame_idx_pairs)
    if data is None:
        return model
    else:
        diff = model - data
        if eps is None:
            return diff
        else:
            return diff / eps


def model_serial_2tcm_frame_avgd(params: lmfit.Parameters,
                                 cp_times: np.ndarray,
                                 cp_vals: np.ndarray,
                                 frame_idx_pairs: np.ndarray,
                                 data: np.ndarray | None = None,
                                 eps: np.ndarray | float | None = None) -> np.ndarray:
    par_vals = params.valuesdict()
    model = gen_tac_2tcm_cpet_from_tac(tac_times=cp_times,
                                       tac_vals=cp_vals,
                                       k1=par_vals['k1'],
                                       k2=par_vals['k2'],
                                       k3=par_vals['k3'],
                                       k4=par_vals['k4'],
                                       vb=par_vals['vb'])[-1]

    model = get_frame_averaged_tac_vals(tac_vals=model, frame_idx_pairs=frame_idx_pairs)
    if data is None:
        return model
    else:
        diff = model - data
        if eps is None:
            return diff
        else:
            return diff / eps